""" test dryrun, that PyGemini can correctly invoke Gemini3D """

import pytest
from pathlib import Path

import gemini3d.run
import gemini3d.job as job
import gemini3d.web

R = Path(__file__).parents[1] / "tests/data"


@pytest.mark.parametrize("name,bref", [("2dew_eq", 1238112), ("3d_eq", 2323072)])
def test_memory(name, bref):

    try:
        ref = gemini3d.web.download_and_extract(name, R)
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    est = job.memory_estimate(ref)

    assert est == bref


def test_mpiexec():

    exe = job.get_gemini_exe()
    assert isinstance(exe, Path)

    # It's OK if MPIexec doesn't exist, but make the test assert consistent with that
    # there are numerous possibilities that MPIexec might not work
    # predicting the outcome of this test requires the function we're testing!
    mpiexec = job.check_mpiexec("mpiexec", exe)
    assert isinstance(mpiexec, str) or mpiexec is None


@pytest.mark.parametrize("name", ["2dew_eq"])
def test_dryrun(name, tmp_path):

    try:
        ref = gemini3d.web.download_and_extract(name, R)
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    params = {
        "config_file": ref,
        "out_dir": tmp_path,
        "dryrun": True,
    }

    job.runner(params)



"""
NetCDF4 file reading
"""

from __future__ import annotations
import xarray
from pathlib import Path
import typing as T
import numpy as np
import logging
from datetime import datetime, timedelta

from .. import find
from .. import WAVELEN

try:
    from netCDF4 import Dataset
except ImportError:
    # must be ImportError not ModuleNotFoundError for botched NetCDF4 linkage
    Dataset = None


def simsize(path: Path) -> tuple[int, ...]:
    """
    get simulation size
    """

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    path = find.simsize(path, ".nc")

    with Dataset(path, "r") as f:
        if "lxs" in f.variables:
            lxs = f["lxs"][:]
        elif "lx" in f.variables:
            lxs = f["lx"][:]
        elif "lx1" in f.variables:
            if f["lx1"].ndim > 0:
                lxs = np.array(
                    [
                        f["lx1"][:].squeeze()[()],
                        f["lx2"][:].squeeze()[()],
                        f["lx3"][:].squeeze()[()],
                    ]
                )
            else:
                lxs = np.array([f["lx1"][()], f["lx2"][()], f["lx3"][()]])
        else:
            raise KeyError(f"could not find 'lxs', 'lx' or 'lx1' in {path.as_posix()}")

    return lxs


def flagoutput(file: Path, cfg: dict[str, T.Any]) -> int:
    """ detect output type """

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    flag = None
    with Dataset(file, "r") as f:
        if "nsall" in f.variables:
            # milestone or full
            flag = 1
        elif "flagoutput" in f.variables:
            flag = f["flagoutput"][()]
        elif "ne" in f.variables and f["ne"].ndim == 3:
            flag = 0
        elif "Tavgall" in f.variables:
            flag = 2
        elif "neall" in f.variables:
            flag = 3
    if flag is None:
        flag = cfg.get("flagoutput")

    return flag


def grid(
    file: Path, *, var: tuple[str, ...] | list[str] = None, shape: bool = False
) -> dict[str, np.ndarray]:
    """
    get simulation grid

    Parameters
    ----------
    file: pathlib.Path
        filepath to simgrid
    var: list of str, optional
        read only these grid variables
    shape: bool, optional
        read only the shape of the grid instead of the data iteslf

    Returns
    -------
    grid: dict
        grid parameters
    """

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    grid: dict[str, T.Any] = {}

    if not file.is_file():
        file2 = find.grid(file)
        if file2 and file2.is_file():
            file = file2
        else:
            logging.error(f"{file} grid file is not present.")
            return grid

    if shape:
        with Dataset(file, "r") as f:
            for key in f.variables:
                grid[key] = f[key].shape

        grid["lxs"] = np.array([grid["x1"], grid["x2"], grid["x3"]])
        return grid

    with Dataset(file, "r") as f:
        if not var:
            var = f.variables
        for k in var:
            if f[k].ndim >= 2:
                grid[k] = f[k][:].transpose()
            else:
                grid[k] = f[k][:]

    grid["lxs"] = simsize(file.with_name("simsize.nc"))

    return grid


def Efield(file: Path) -> xarray.Dataset:
    """
    load electric field
    """

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    with Dataset(file.with_name("simgrid.nc"), "r") as f:
        E = xarray.Dataset(coords={"mlon": f["/mlon"][:], "mlat": f["/mlat"][:]})

    with Dataset(file, "r") as f:
        E["flagdirich"] = f["flagdirich"][()].item()
        for p in ("Exit", "Eyit", "Vminx1it", "Vmaxx1it"):
            E[p] = (("mlat", "mlon"), f[p][:])
        for p in ("Vminx2ist", "Vmaxx2ist"):
            E[p] = (("mlat",), f[p][:])
        for p in ("Vminx3ist", "Vmaxx3ist"):
            E[p] = (("mlon",), f[p][:])

    return E


def precip(file: Path) -> xarray.Dataset:
    """
    load precipitation
    """

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    with Dataset(file.with_name("simgrid.nc"), "r") as f:
        dat = xarray.Dataset(coords={"mlon": f["/mlon"][:], "mlat": f["/mlat"][:]})

    with Dataset(file, "r") as f:
        for k in ("Q", "E0"):
            dat[k] = (("mlat", "mlon"), f[f"/{k}p"][:])

    return dat


def frame3d_curvne(file: Path) -> xarray.Dataset:
    """
    just Ne
    """

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    xg = grid(file.parent, var=("x1", "x2", "x3"))
    dat = xarray.Dataset(coords={"x1": xg["x1"][2:-2], "x2": xg["x2"][2:-2], "x3": xg["x3"][2:-2]})

    with Dataset(file, "r") as f:
        dat["ne"] = (("x1", "x2", "x3"), f["/ne"][:])

    return dat


def frame3d_curv(file: Path, var: tuple[str, ...] | list[str]) -> xarray.Dataset:
    """
    curvilinear

    Parameters
    ----------

    file: pathlib.Path
        filename to read
    var: list of str
        variable(s) to read
    """

    xg = grid(file.parent, var=("x1", "x2", "x3"))
    dat = xarray.Dataset(coords={"x1": xg["x1"][2:-2], "x2": xg["x2"][2:-2], "x3": xg["x3"][2:-2]})

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    lxs = simsize(file.parent)

    p4 = (0, 3, 2, 1)
    if lxs[2] == 1:  # east-west
        p3 = (2, 0, 1)
    else:  # 3D or north-south, no swap
        p3 = (2, 1, 0)

    with Dataset(file, "r") as f:
        if {"ne", "ns", "v1", "Ti"}.intersection(var):
            dat["ns"] = (("species", "x1", "x2", "x3"), f["nsall"][:].transpose(p4))

        if {"v1", "vs1"}.intersection(var):
            dat["vs1"] = (("species", "x1", "x2", "x3"), f["vs1all"][:].transpose(p4))

        if {"Te", "Ti", "Ts"}.intersection(var):
            dat["Ts"] = (("species", "x1", "x2", "x3"), f["Tsall"][:].transpose(p4))

        for k in {"J1", "J2", "J3"}.intersection(var):
            dat[k] = (("x1", "x2", "x3"), f[f"{k}all"][:].transpose(p3))

        for k in {"v2", "v3"}.intersection(var):
            dat[k] = (("x1", "x2", "x3"), f[f"/{k}avgall"][:].transpose(p3))

        if "Phi" in var:
            Phiall = f["Phiall"][:]
            if Phiall.ndim == 2:
                Phiall = Phiall.transpose()
            elif Phiall.ndim == 1:
                if dat.x2.size == 1:
                    Phiall = Phiall[:, None]
                else:
                    Phiall = Phiall[None, :]

            dat["Phitop"] = (("x2", "x3"), Phiall)

    return dat


def frame3d_curvavg(file: Path, var: tuple[str, ...] | list[str]) -> xarray.Dataset:
    """

    Parameters
    ----------
    file: pathlib.Path
        filename of this timestep of simulation output
    var: list of str
        variable(s) to read
    """

    xg = grid(file.parent, var=("x1", "x2", "x3"))
    dat = xarray.Dataset(coords={"x1": xg["x1"][2:-2], "x2": xg["x2"][2:-2], "x3": xg["x3"][2:-2]})

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    lxs = simsize(file.parent)

    if lxs[2] == 1:  # east-west
        p3 = (2, 0, 1)
    else:  # 3D or north-south, no swap
        p3 = (2, 1, 0)

    v2n = {
        "ne": "neall",
        "v1": "v1avgall",
        "Ti": "Tavgall",
        "Te": "TEall",
        "J1": "J1all",
        "J2": "J2all",
        "J3": "J3all",
        "v2": "v2avgall",
        "v3": "v3avgall",
        "Phi": "Phiall",
    }

    with Dataset(file, "r") as f:
        for k in var:
            if k == "Phi":
                dat["Phitop"] = (("x2", "x3"), f[v2n[k]][:].transpose())
            else:
                dat[k] = (("x1", "x2", "x3"), f[v2n[k]][:].transpose(p3))

    return dat


def glow_aurmap(file: Path) -> xarray.Dataset:
    """
    read the auroral output from GLOW

    Parameters
    ----------
    file: pathlib.Path
        filename of this timestep of simulation output
    """

    xg = grid(file.parents[1], var=("x2", "x3"))
    dat = xarray.Dataset(coords={"wavelength": WAVELEN, "x2": xg["x2"][2:-2], "x3": xg["x3"][2:-2]})

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    with Dataset(file, "r") as h:
        dat["rayleighs"] = (("wavelength", "x2", "x3"), h["iverout"][:])

    return dat


def time(file: Path) -> datetime:
    """
    reads simulation time
    """

    if Dataset is None:
        raise ImportError("netcdf missing or broken")

    with Dataset(file, "r") as f:
        ymd = datetime(*f["ymd"][:3])

        if "UThour" in f:
            hour = f["UThour"][()].item()
        elif "UTsec" in f:
            hour = f["UTsec"][()].item() / 3600
        else:
            raise KeyError(f"did not find time of day in {file}")

    t = ymd + timedelta(hours=hour)

    return t



from matplotlib.figure import Figure
import numpy as np
from pathlib import Path
from datetime import datetime
import logging


def plotdiff(
    A: np.ndarray,
    B: np.ndarray,
    name: str,
    time: datetime,
    outdir: Path,
    refdir: Path,
):

    A = A.squeeze()
    B = B.squeeze()

    if A.ndim == 3:
        # loop over the species, which are in the first dimension
        for i in range(A.shape[0]):
            plotdiff(A[i], B[i], f"{name}-{i}", time, outdir, refdir)

    if A.ndim not in (1, 2):
        logging.error(f"skipping diff plot: {name}")
        return None

    fg = Figure(constrained_layout=True, figsize=(12, 5))
    axs = fg.subplots(1, 3)

    if A.ndim == 2:
        diff2d(A, B, name, fg, axs)
    elif A.ndim == 1:
        diff1d(A, B, name, fg, axs)

    axs[0].set_title(str(outdir))
    axs[1].set_title(str(refdir))
    axs[2].set_title(f"diff: {name}")

    tstr = time.isoformat()
    ttxt = f"{name}  {tstr}"

    fg.suptitle(ttxt)

    fn = outdir / f"{name}-diff-{tstr.replace(':','')}.png"
    print("writing", fn)
    fg.savefig(fn)


def diff1d(A: np.ndarray, B: np.ndarray, name: str, fg, axs):

    axs[0].plot(A)

    axs[1].plot(B)

    axs[2].plot(A - B)


def diff2d(A: np.ndarray, B: np.ndarray, name: str, fg, axs):

    cmap = "bwr" if name.startswith(("J", "v")) else None

    bmin = min(A.min(), B.min())
    bmax = max(A.max(), B.max())

    hi = axs[0].pcolormesh(A, cmap=cmap, vmin=bmin, vmax=bmax)
    fg.colorbar(hi, ax=axs[0])

    hi = axs[1].pcolormesh(B, cmap=cmap, vmin=bmin, vmax=bmax)
    fg.colorbar(hi, ax=axs[1])

    dAB = A - B
    b = max(abs(dAB.min()), abs(dAB.max()))

    hi = axs[2].pcolormesh(A - B, cmap="bwr", vmin=-b, vmax=b)
    fg.colorbar(hi, ax=axs[2])



from __future__ import annotations
import numpy as np
import xarray
from pathlib import Path
import typing as T
import sys

from .utils import git_meta
from .hdf5 import write as h5write
from .nc4 import write as ncwrite
from . import namelist


def state(out_file: Path, dat: xarray.Dataset, file_format: str = None, **kwargs):
    """
    WRITE STATE VARIABLE DATA.
    NOTE: WE don't write ANY OF THE ELECTRODYNAMIC
    VARIABLES SINCE THEY ARE NOT NEEDED TO START THINGS
    UP IN THE FORTRAN CODE.

    INPUT ARRAYS SHOULD BE TRIMMED TO THE CORRECT SIZE
    I.E. THEY SHOULD NOT INCLUDE GHOST CELLS
    """

    ext = file_format if file_format else out_file.suffix[1:]

    # %% allow overriding "dat"
    if "time" in kwargs:
        dat.attrs["time"] = kwargs["time"]

    for k in {"ns", "vs1", "Ts"}:
        if k in kwargs:
            dat[k] = (("species", "x1", "x2", "x3"), kwargs[k])

    if "Phitop" in kwargs:
        dat["Phitop"] = (("x2", "x3"), kwargs["Phitop"])

    # %% dispatch to format-specific writers
    if ext == "h5":
        h5write.state(out_file.with_suffix(".h5"), dat)
    elif ext == "nc":
        ncwrite.state(out_file.with_suffix(".nc"), dat)
    else:
        raise ValueError(f"unknown file format {ext}")


def data(out_file: Path, dat: np.ndarray, file_format: str, xg: dict[str, T.Any] = None):

    if file_format == "h5":
        h5write.data(out_file, dat)
    elif file_format == "nc":
        ncwrite.data(out_file, dat, xg)
    else:
        raise ValueError(f"Unknown file format {file_format}")


def grid(cfg: dict[str, T.Any], xg: dict[str, T.Any], *, file_format: str = None):
    """writes grid to disk

    Parameters
    ----------

    cfg: dict
        simulation parameters
    xg: dict
        grid values

    NOTE: we use .with_suffix() in case file_format was overriden by user
    that allows writing NetCDF4 and HDF5 by scripts using same input files
    """

    input_dir = cfg["indat_size"].parent
    if input_dir.is_file():
        raise OSError(f"{input_dir} is a file instead of directory")

    input_dir.mkdir(parents=True, exist_ok=True)

    if not file_format:
        file_format = cfg["file_format"] if "file_format" in cfg else cfg["indat_size"].suffix[1:]

    if file_format == "h5":
        h5write.grid(cfg["indat_size"].with_suffix(".h5"), cfg["indat_grid"].with_suffix(".h5"), xg)
    elif file_format == "nc":
        ncwrite.grid(cfg["indat_size"].with_suffix(".nc"), cfg["indat_grid"].with_suffix(".nc"), xg)
    else:
        raise ValueError(f'unknown file format {cfg["file_format"]}')

    meta(cfg["out_dir"] / "setup_meta.nml", git_meta(), "setup_python")


def Efield(E: xarray.Dataset, outdir: Path, file_format: str):
    """writes E-field to disk

    Parameters
    ----------

    E: dict
        E-field values
    outdir: pathlib.Path
        directory to write files into
    file_format: str
        requested file format to write
    """

    print("write E-field data to", outdir)
    outdir.mkdir(parents=True, exist_ok=True)

    if file_format == "h5":
        h5write.Efield(outdir, E)
    elif file_format == "nc":
        ncwrite.Efield(outdir, E)
    else:
        raise ValueError(f"unknown file format {file_format}")


def precip(precip: dict[str, T.Any], outdir: Path, file_format: str):
    """writes precipitation to disk

    Parameters
    ----------
    precip: dict
        preicipitation values
    outdir: pathlib.Path
        directory to write files into
    file_format: str
        requested file format to write
    """

    print("write precipitation data to", outdir)
    outdir.mkdir(parents=True, exist_ok=True)

    if file_format == "h5":
        h5write.precip(outdir, precip)
    elif file_format == "nc":
        ncwrite.precip(outdir, precip)
    else:
        raise ValueError(f"unknown file format {file_format}")


def meta(fn: Path, meta: dict[str, str], nml: str):
    """
    writes Namelist file with metadata
    """

    fn = fn.expanduser()
    if fn.is_dir():
        fn = fn / "setup_meta.nml"

    meta["python_version"] = sys.version

    namelist.write(fn, nml, meta)



"""
read and write Fortran standard namelist
This is very basic--see f90nml Python package.
"""

from __future__ import annotations
import typing as T
from pathlib import Path
import re
import numpy as np


__all__ = ["read", "write"]


def read(file: Path, namelist: str) -> dict[str, T.Any]:
    """read a namelist from an .nml file, as strings

    Parameters
    ----------

    file: pathlib.Path
        Namelist file to read
    namelist: str
        Namelist to read from file

    Returns
    -------

    cfg: dict
        data contained in namelist
    """

    r: dict[str, list[str]] = {}
    nml_pat = re.compile(r"^\s*&(" + namelist + r")")
    end_pat = re.compile(r"^\s*/\s*$")
    val_pat = re.compile(r"^\s*(\w+)\s*=\s*([^!]*)")

    with file.open("r") as f:
        for line in f:
            if not nml_pat.match(line):
                continue

            for line in f:
                if end_pat.match(line):
                    # end of namelist
                    return r
                val_mat = val_pat.match(line)
                if not val_mat:
                    continue

                key, vals = val_mat.group(1), val_mat.group(2).strip().split(",")
                values: list[T.Any] = []
                for v in vals:
                    v = v.strip().replace("'", "").replace('"', "")
                    try:
                        values.append(float(v))
                    except ValueError:
                        values.append(v)
                r[key] = values[0] if len(values) == 1 else values

    raise KeyError(f"did not find Namelist {namelist} in {file}")


def write(file: Path, namelist: str, data: dict[str, T.Any], overwrite: bool = False):
    """
    writes a basic Fortran namelist to a .nml text file

    Parameters
    ----------

    file: pathlib.Path
        Namelist file to read
    namelist: str
        Namelist to read from file
    data: dict
        data to write to namelist
    """

    def _write_scalar(f, key: str, value: T.Any):
        f.write(f"{key} = {value}\n")

    def _write_string(f, key: str, value: str):
        f.write(f'{key} = "{value}"\n')

    def _write_value(f, key: str, value: T.Any):
        if isinstance(value, (float, int)):
            _write_scalar(f, key, value)
        elif isinstance(value, str):
            _write_string(f, key, value)
        elif isinstance(value, (tuple, list)):
            if isinstance(value[0], str):
                s = ",".join([f'"{v}"' for v in value])
                _write_scalar(f, key, s)
                return
            s = ",".join(map(str, value))
            _write_scalar(f, key, s)
        elif isinstance(value, np.ndarray):
            _write_scalar(f, key, value.astype(str))
        else:
            raise TypeError(f"unsure how to handle {type(value)}")

    file = file.expanduser().resolve()

    if file.is_dir():
        raise OSError(f"give a filename, not a directory {file}")

    mode = "w" if overwrite else "a"

    with file.open(mode=mode) as f:
        if mode == "a":
            # arbitrary, for aesthetics
            f.write("\n")

        f.write(f"&{namelist}\n")

        for k, v in data.items():
            _write_value(f, k, v)

        f.write("/\n")
        # closes namelist
        # ensure a trailing blank line or file will fail to read in Fortran



"""
plasma functions
"""

from __future__ import annotations
import typing as T
import numpy as np
import logging
import xarray
from scipy.integrate import cumtrapz
from scipy.interpolate import interp1d, interp2d, interpn

from . import read
from . import LSP, SPECIES
from . import write
from .web import url_retrieve, extract_zip
from .msis import msis_setup

# CONSTANTS
KB = 1.38e-23
AMU = 1.67e-27


def equilibrium_resample(p: dict[str, T.Any], xg: dict[str, T.Any]):
    """
    read and interpolate equilibrium simulation data, writing new
    interpolated grid.
    """

    # %% download equilibrium data if needed and specified
    if not p["eq_dir"].is_dir():
        if "eq_url" not in p:
            raise FileNotFoundError(
                f"{p['eq_dir']} not found and eq_url not specified in {p['nml']}"
            )
        url_retrieve(p["eq_url"], p["eq_zip"])
        extract_zip(p["eq_zip"], p["eq_dir"])

    # %% READ Equilibrium SIMULATION INFO
    peq = read.config(p["eq_dir"])
    if not peq:
        raise FileNotFoundError(
            f"equilibrium directory {p['eq_dir']} does not appear to contain config.nml"
        )

    # %% END FRAME time of equilibrium simulation
    # this will be the starting time of the new simulation
    t_eq_end = peq["time"][-1]

    # %% LOAD THE last equilibrium frame
    dat = read.frame(p["eq_dir"], t_eq_end)
    if not dat:
        raise FileNotFoundError(f"{p['eq_dir']} does not have data for {t_eq_end}")

    # %% sanity check equilibrium simulation input to interpolation
    check_density(dat["ns"])
    check_drift(dat["vs1"])
    check_temperature(dat["Ts"])

    # %% DO THE INTERPOLATION
    xg_in = read.grid(p["eq_dir"])
    if not xg_in:
        raise FileNotFoundError(f"{p['eq_dir']} does not have an input simulation grid.")

    dat_interp = model_resample(xg_in, dat, xg)

    # %% sanity check interpolated variables
    check_density(dat_interp["ns"])
    check_drift(dat_interp["vs1"])
    check_temperature(dat_interp["Ts"])

    # %% WRITE OUT THE GRID
    write.grid(p, xg)

    write.state(p["indat_file"], dat_interp, file_format=p.get("file_format"))


def model_resample(
    xgin: dict[str, T.Any], dat: xarray.Dataset, xg: dict[str, T.Any]
) -> xarray.Dataset:
    """resample a grid
    usually used to upsample an equilibrium simulation grid

    Parameters
    ----------

    xgin: dict
        original grid (usually equilibrium sim grid)
    dat: xarray.Dataset
        data to interpolate

    Returns
    -------

    dat_interp: xarray.Dataset
        interpolated data
    """

    # %% NEW GRID SIZES
    lx1, lx2, lx3 = xg["lx"]

    # %% ALLOCATIONS

    dat_interp = xarray.Dataset(
        coords={
            "species": SPECIES,
            "x1": xg["x1"][2:-2],
            "x2": xg["x2"][2:-2],
            "x3": xg["x3"][2:-2],
        }
    )
    for k in {"ns", "vs1", "Ts"}:
        dat_interp[k] = (
            ("species", "x1", "x2", "x3"),
            np.empty((LSP, lx1, lx2, lx3), dtype=np.float32),
        )

    # %% INTERPOLATE ONTO NEWER GRID
    # to avoid IEEE754 rounding issues leading to bounds error,
    # cast the arrays to the same precision,
    # preferring float32 to save disk space and IO time
    X2 = xgin["x2"][2:-2].astype(np.float32)
    X1 = xgin["x1"][2:-2].astype(np.float32)
    X3 = xgin["x3"][2:-2].astype(np.float32)
    x1i = xg["x1"][2:-2].astype(np.float32)
    x2i = xg["x2"][2:-2].astype(np.float32)
    x3i = xg["x3"][2:-2].astype(np.float32)

    if lx3 > 1 and lx2 > 1:
        # 3-D
        logging.info("interpolating grid for 3-D simulation")
        # X2, X1, X3 = np.meshgrid(xgin['x2'][2:-2], xgin['x1'][2:-2], xgin['x3'][2:-2])
        X2i, X1i, X3i = np.meshgrid(x2i, x1i, x3i)
        assert X2i.shape == tuple(xg["lx"])

        for i in range(LSP):
            for k in {"ns", "vs1", "Ts"}:
                # the .values are to avoid OutOfMemoryError
                dat_interp[k][i, :, :, :] = interpn(
                    points=(X1, X2, X3),
                    values=dat[k][i, :, :, :].values,
                    xi=(X1i, X2i, X3i),
                    bounds_error=True,
                )

    elif lx3 == 1:
        # 2-D east-west
        logging.info("interpolating grid for 2-D simulation in x1, x2")
        # [X2,X1]=meshgrid(xgin.x2(3:end-2),xgin.x1(3:end-2));
        # [X2i,X1i]=meshgrid(xg.x2(3:end-2),xg.x1(3:end-2));
        for i in range(LSP):
            for k in {"ns", "vs1", "Ts"}:
                f = interp2d(X2, X1, dat[k][i, :, :, :], bounds_error=True)
                dat_interp[k][i, :, :, :] = f(x2i, x1i)[:, :, None]

    elif lx2 == 1:
        # 2-D north-south
        logging.info("interpolating grid for 2-D simulation in x1, x3")
        # original grid, a priori the first 2 and last 2 values are ghost cells
        # on each axis
        #
        # Detect old non-padded grid and workaround
        if np.isclose(xgin["x3"][0], xg["x3"][2], atol=1):
            # old sim, no external ghost cells.
            # Instead of discarding good cells,keep them and say there are
            # new ghost cells outside the grid
            X3 = np.linspace(xgin["x3"][0], xgin["x3"][-1], xgin["lx"][2])
        else:
            # new sim, external ghost cells
            X3 = xgin["x3"][2:-2]

        X1 = xgin["x1"][2:-2]
        # new grid
        x3i = xg["x3"][2:-2].astype(np.float32)
        x1i = xg["x1"][2:-2].astype(np.float32)

        # for each species
        for i in range(LSP):
            for k in {"ns", "vs1", "Ts"}:
                f = interp2d(X3, X1, dat[k][i, :, :, :], bounds_error=True)
                dat_interp[k][i, :, :, :] = f(x3i, x1i)[:, None, :]

    else:
        raise ValueError("Not sure if this is 2-D or 3-D simulation")

    dat_interp.attrs["time"] = dat.time

    return dat_interp


def check_density(n: np.ndarray):

    if not np.isfinite(n).all():
        raise ValueError("non-finite density")
    if (n < 0).any():
        raise ValueError("negative density")
    if n.max() < 1e6:
        raise ValueError("too small maximum density")


def check_drift(v: np.ndarray):

    if not np.isfinite(v).all():
        raise ValueError("non-finite drift")
    if (abs(v) > 10e3).any():
        raise ValueError("excessive drift velocity")


def check_temperature(Ts: np.ndarray):

    if not np.isfinite(Ts).all():
        raise ValueError("non-finite temperature")
    if (Ts < 0).any():
        raise ValueError("negative temperature")
    if Ts.max() < 500:
        raise ValueError("too cold maximum temperature")


def equilibrium_state(p: dict[str, T.Any], xg: dict[str, T.Any]) -> xarray.Dataset:
    """
    generate (arbitrary) initial conditions for a grid.
    NOTE: only works on symmmetric closed grids!

    [f107a, f107, ap] = activ
    """

    # %% MAKE UP SOME INITIAL CONDITIONS FOR FORTRAN CODE
    mindens = 1e-100

    def Oplus(ns: np.ndarray) -> np.ndarray:
        ns[0, :, ix2, ix3] = rho * ne
        zref = 900e3
        i = alt[:, ix2, ix3] > zref
        if any(i):
            iord = np.argsort(alt[:, ix2, ix3])
            altsort = alt[iord, ix2, ix3]
            nsort = ns[0, :, ix2, ix3]
            nsort = nsort[iord]

            ms = 16 * AMU
            H = KB * 2 * Tn[inds, ix2, ix3] / ms / g[inds, ix2, ix3]
            z = alt[i, ix2, ix3]
            lz = z.size
            iord = np.argsort(z)
            z = z[iord]
            #     z=[z; 2*z(lz)-z(lz-1)];
            z = np.insert(z, 0, zref)
            integrand = 1 / H[iord]
            integrand = np.append(integrand, integrand[-1])

            # this cumtrapz() does NOT get initial=0, since matlab user code strips first element here
            redheight = cumtrapz(integrand, z)
            f = interp1d(altsort, nsort)
            n1top = f(zref) * np.exp(-redheight)
            n1sort = np.zeros(lz)
            for iz in range(lz):
                n1sort[iord[iz]] = n1top[iz]

            ns[0, i, ix2, ix3] = n1sort

        return ns

    def molecular_density(ns: np.ndarray, xgr: np.ndarray, inds: np.ndarray) -> np.ndarray:
        """MOLECULAR DENSITIES

        Parameters
        ----------

        ns: np.ndarray
            4D by species number density
        xgr: np.ndarray
            xg["r"]
        inds: np.ndarray
            boolean vector

        Returns
        -------

        ns: np.ndarray
            4D by species number density

        """

        i = np.setdiff1d(range(lx1), inds.nonzero()[0])

        nmolc = np.zeros(lx1)
        nmolc[i] = (1 - rho[i]) * ne[i]

        cond: bool = None

        if any(inds):
            if xgr.ndim == 3:
                cond = xgr[0, 0, 0] > xgr[1, 0, 0]
            elif xgr.ndim == 2:
                cond = xgr[0, 0] > xgr[1, 0]
            else:
                raise ValueError(
                    "xg['r'] expected to be 3D, possibly with degenerate 2nd or 3rd dimension"
                )

            iref = i[0] if cond else i[-1]

            n0 = nmolc[iref]
            ms = 30.5 * AMU
            H = KB * Tn[inds, ix2, ix3] / ms / g[inds, ix2, ix3]
            z = alt[inds, ix2, ix3]
            lz = z.size
            iord = np.argsort(z)
            z = z[iord]
            z = np.append(z, 2 * z[-1] - z[-2])
            integrand = 1 / H[iord]
            integrand = np.append(integrand, integrand[-1])
            # this cumtrapz() does NOT get initial=0, since matlab user code strips first element here
            redheight = cumtrapz(integrand, z)
            nmolctop = n0 * np.exp(-redheight)
            nmolcsort = np.zeros(lz)
            for iz in range(lz):
                nmolcsort[iord[iz]] = nmolctop[iz]

            nmolc[inds] = nmolcsort

        ns[1, :, ix2, ix3] = 1 / 3 * nmolc
        ns[2, :, ix2, ix3] = 1 / 3 * nmolc
        ns[3, :, ix2, ix3] = 1 / 3 * nmolc

        # %% PROTONS
        ns[5, inds, ix2, ix3] = (1 - rho[inds]) * ne[inds]
        z = alt[i, ix2, ix3]
        if any(inds):
            iref = inds.nonzero()[0][-1] if cond else inds.nonzero()[0][0]
            n0 = ns[5, iref, ix2, ix3]
        else:
            iref = alt[:, ix2, ix3].argmax()
            n0 = 1e6

        ns[5, i, ix2, ix3] = chapmana(z, n0, alt[iref, ix2, ix3], Hf.mean().item())

        return ns

    # %% SLICE THE FIELD IN HALF IF WE ARE CLOSED
    atmos = msis_setup(p, xg)

    closeddip = abs(xg["r"][0, 0, 0] - xg["r"][-1, 0, 0]) < 50e3
    # logical flag marking the grid as closed dipole
    if closeddip:
        # closed dipole grid
        #    [~,ialtmax]=max(xg.alt(:,1,1))
        #    lalt=ialtmax
        lalt = xg["lx"][0] // 2
        # FIXME:  needs to work with asymmetric grid...
        alt = xg["alt"][:lalt, :, :]
        lx1 = lalt
        lx2 = xg["lx"][1]
        lx3 = xg["lx"][2]
        Tn = atmos["Tn"][:lalt, :, :]
        g = abs(xg["gx1"][:lalt, :, :])
        g = max(g, 1)
        for ix3 in range(lx3):
            for ix2 in range(lx2):
                ialt = abs(g[:, ix2, ix3] - 1).argmin()
                if ialt != lalt:
                    g[ialt:lalt, ix2, ix3] = 1

    else:
        alt = xg["alt"]
        lx1, lx2, lx3 = xg["lx"]
        Tn = atmos["Tn"]
        g = abs(xg["gx1"])

    ns = np.zeros((7, lx1, lx2, lx3), dtype=np.float32)
    for ix3 in range(lx3):
        for ix2 in range(lx2):
            Hf = KB * Tn[:, ix2, ix3] / AMU / 16 / g[:, ix2, ix3]
            z0f = 325e3
            He = 2 * KB * Tn[:, ix2, ix3] / AMU / 30 / g[:, ix2, ix3]
            z0e = 120e3
            ne = chapmana(alt[:, ix2, ix3], p["nmf"], z0f, Hf) + chapmana(
                alt[:, ix2, ix3], p["nme"], z0e, He
            )
            rho = 1 / 2 * np.tanh((alt[:, ix2, ix3] - 200e3) / 45e3) - 1 / 2 * np.tanh(
                (alt[:, ix2, ix3] - 1000e3) / 200e3
            )

            inds = alt[:, ix2, ix3] > z0f
            if any(inds):
                ms = rho[inds] * 16 * AMU + (1 - rho[inds]) * AMU
                # topside composition only
                H = KB * 2 * Tn[inds, ix2, ix3] / ms / g[inds, ix2, ix3]
                z = alt[inds, ix2, ix3]
                lz = z.size
                iord = np.argsort(z)
                z = z[iord]
                #     z=[z; 2*z(lz)-z(lz-1)];
                z = np.insert(z, 0, z0f)
                integrand = 1 / H[iord]
                integrand = np.append(integrand, integrand[-1])
                # initial=0 is to match Matlab cumtrapz()
                redheight = cumtrapz(integrand, z, initial=0)
                netop = p["nmf"] * np.exp(-redheight)
                nesort = np.zeros(lz)
                for iz in range(lz):
                    nesort[iord[iz]] = netop[iz]

                ne[inds] = nesort

            ns = Oplus(ns)

            # N+
            ns[4, :, ix2, ix3] = 1e-4 * ns[0, :, ix2, ix3]

            ns = molecular_density(ns, xg["r"], inds)

    ns[:6, :, :, :][ns[:6, :, :, :] < mindens] = mindens
    ns[6, :, :, :] = ns[:6, :, :, :].sum(axis=0)

    vsx1 = np.zeros((7, lx1, lx2, lx3), dtype=np.float32)

    Ts = np.broadcast_to(Tn, [7, lx1, lx2, lx3])

    if closeddip:
        # closed dipole grid
        # FIXME:  This code only works for symmetric grids...
        if 2 * lx1 == xg["lx"][0]:
            ns = np.concatenate((ns, ns[:, ::-1, :, :]), 1)
            Ts = np.concatenate((Ts, Ts[:, ::-1, :, :]), 1)
            vsx1 = np.concatenate((vsx1, vsx1[:, ::-1, :, :]), 1)
        else:
            ns = np.concatenate((ns, ns[:, lx1, :, :], ns[:, ::-1, :, :]), 1)
            Ts = np.concatenate((Ts, Ts[:, lx1, :, :], Ts[:, ::-1, :, :]), 1)
            vsx1 = np.concatenate((vsx1, vsx1[:, lx1, :, :], vsx1[:, ::-1, :, :]), 1)

    dat = xarray.Dataset(
        {
            "ns": (("species", "x1", "x2", "x3"), ns),
            "vs1": (("species", "x1", "x2", "x3"), vsx1),
            "Ts": (("species", "x1", "x2", "x3"), Ts),
        },
        coords={
            "species": SPECIES,
            "x1": xg["x1"][2:-2],
            "x2": xg["x2"][2:-2],
            "x3": xg["x3"][2:-2],
        },
        attrs={"time": p["time"][0]},
    )

    return dat


def chapmana(z: np.ndarray, nm: float, z0: float, H: float) -> np.ndarray:
    zref = (z - z0) / H
    ne = nm * np.exp(0.5 * (1 - zref - np.exp(-zref)))

    ne[ne < 1] = 1

    return ne



"""
NetCDF4 file writing
"""

from __future__ import annotations
import xarray
import typing as T
import numpy as np
from datetime import datetime
from pathlib import Path
import logging

from ..utils import datetime2ymd_hourdec, to_datetime
from .. import LSP

try:
    from netCDF4 import Dataset
except ImportError:
    # must be ImportError not ModuleNotFoundError for botched NetCDF4 linkage
    Dataset = None


def state(fn: Path, dat: xarray.Dataset):
    """
     WRITE STATE VARIABLE DATA.
    NOTE THAT WE don't write ANY OF THE ELECTRODYNAMIC
    VARIABLES SINCE THEY ARE NOT NEEDED TO START THINGS
    UP IN THE FORTRAN CODE.

    INPUT ARRAYS SHOULD BE TRIMMED TO THE CORRECT SIZE
    I.E. THEY SHOULD NOT INCLUDE GHOST CELLS
    """

    if Dataset is None:
        raise ImportError("pip install netcdf4")

    time = dat.time

    logging.info(f"state: {fn} {time}")

    with Dataset(fn, "w") as f:
        f.createDimension("ymd", 3)
        g = f.createVariable("ymd", np.int32, "ymd")
        g[:] = [time.year, time.month, time.day]

        g = f.createVariable("UTsec", np.float32)
        g[:] = time.hour * 3600 + time.minute * 60 + time.second + time.microsecond / 1e6
        f.createDimension("species", 7)
        f.createDimension("x1", dat.x1.size)
        f.createDimension("x2", dat.x2.size)
        f.createDimension("x3", dat.x3.size)

        for k in {"ns", "vs1", "Ts"}:
            if k in dat.data_vars:
                _write_var(f, f"{k}all", dat[k])

        if "Phitop" in dat.data_vars:
            _write_var(f, "Phiall", dat["Phitop"])


def data(fn: Path, dat: xarray.Dataset, xg: dict[str, T.Any]):
    """
    write simulation data
    e.g. for converting a file format from a simulation
    """

    if Dataset is None:
        raise ImportError("pip install netcdf4")

    logging.info(f"write_data: {fn}")

    if "ns" in dat:
        shape = dat["ns"].shape
    elif "ne" in dat:
        shape = dat["ne"].shape
    else:
        raise ValueError("what variable should I use to determine dimensions?")

    with Dataset(fn, "w") as f:

        if len(shape) == 4:
            dims = ["species", "x1", "x2", "x3"]
            f.createDimension(dims[0], LSP)
            f.createDimension(dims[1], shape[1])
            f.createDimension(dims[2], shape[2])
            f.createDimension(dims[3], shape[3])
        elif len(shape) == 3:
            dims = ["x1", "x2", "x3"]
            f.createDimension(dims[0], shape[0])
            f.createDimension(dims[1], shape[1])
            f.createDimension(dims[2], shape[2])
        else:
            raise ValueError("unknown how to handle non 3-D or 4-D array")

        # set dimension values
        if xg:
            Ng = 4  # number of ghost cells
            for k in dims[-3:]:
                if xg[k].size == f.dimensions[k].size + Ng:
                    # omit ghost cells
                    x = xg[k][2:-2]
                elif xg[k].size == f.dimensions[k].size:
                    x = xg[k]
                else:
                    raise ValueError(f"{k}:  {xg[k].size} != {f.dimensions[k].size}")
                _write_var(f, name=k, value=x, dims=k)

        for k in ["ns", "vs1", "Ts"]:
            if k not in dat:
                continue
            _write_var(f, k, dat[k], dims)

        for k in ["ne", "v1", "Ti", "Te", "J1", "J2", "J3", "v2", "v3"]:
            if k not in dat:
                continue
            _write_var(f, k, dat[k], dims)

        if "Phitop" in dat:
            _write_var(f, "Phiall", dat["Phitop"], dims[-2:])


def grid(size_fn: Path, grid_fn: Path, xg: dict[str, T.Any]):
    """writes grid to disk

    Parameters
    ----------

    size_fn: pathlib.Path
        file to write
    grid_fn: pathlib.Path
        file to write
    xg: dict
        grid values
    """

    if Dataset is None:
        raise ImportError("pip install netcdf4")

    logging.info(f"grid: {size_fn}")

    with Dataset(size_fn, "w") as f:
        f.createDimension("length", len(xg["lx"]))
        g = f.createVariable("lx", np.int32, ("length",))
        g[:] = xg["lx"]

    logging.info(f"write_grid: {grid_fn}")
    Ng = 4  # number of ghost cells

    with Dataset(grid_fn, "w") as f:
        f.createDimension("x1ghost", xg["lx"][0] + Ng)
        f.createDimension("x1d", xg["lx"][0] + Ng - 1)
        f.createDimension("x1i", xg["lx"][0] + 1)
        f.createDimension("x1", xg["lx"][0])

        f.createDimension("x2ghost", xg["lx"][1] + Ng)
        f.createDimension("x2d", xg["lx"][1] + Ng - 1)
        f.createDimension("x2i", xg["lx"][1] + 1)
        f.createDimension("x2", xg["lx"][1])

        f.createDimension("x3ghost", xg["lx"][2] + Ng)
        f.createDimension("x3d", xg["lx"][2] + Ng - 1)
        f.createDimension("x3i", xg["lx"][2] + 1)
        f.createDimension("x3", xg["lx"][2])

        f.createDimension("ecef", 3)

        for i in (1, 2, 3):
            _write_var(f, f"x{i}", xg[f"x{i}"], dims=f"x{i}ghost")
            _write_var(f, f"x{i}i", xg[f"x{i}i"], dims=f"x{i}i")
            _write_var(f, f"dx{i}b", xg[f"dx{i}b"], dims=f"x{i}d")
            _write_var(f, f"dx{i}h", xg[f"dx{i}h"], dims=f"x{i}")
            _write_var(f, f"h{i}", xg[f"h{i}"], ("x1ghost", "x2ghost", "x3ghost"))
            _write_var(f, f"h{i}x1i", xg[f"h{i}x1i"], ("x1i", "x2", "x3"))
            _write_var(f, f"h{i}x2i", xg[f"h{i}x2i"], ("x1", "x2i", "x3"))
            _write_var(f, f"h{i}x3i", xg[f"h{i}x3i"], ("x1", "x2", "x3i"))
            _write_var(f, f"gx{i}", xg[f"gx{i}"], ("x1", "x2", "x3"))
            _write_var(f, f"e{i}", xg[f"e{i}"], ("x1", "x2", "x3", "ecef"))

        for k in ("alt", "glat", "glon", "Bmag", "nullpts", "r", "theta", "phi", "x", "y", "z"):
            _write_var(f, k, xg[k], ("x1", "x2", "x3"))

        for k in ("er", "etheta", "ephi"):
            _write_var(f, k, xg[k], ("x1", "x2", "x3", "ecef"))

        _write_var(f, "I", xg["I"], ("x2", "x3"))


def _write_var(
    f,
    name: str,
    value: np.ndarray | xarray.DataArray,
    dims: str | tuple[str, ...] | list[str] = None,
):

    if dims is None and isinstance(value, xarray.DataArray):
        dims = value.dims
    elif isinstance(dims, str):
        dims = [dims]
    else:
        raise ValueError(f"Please specify dims for {name}")

    dims = dims[::-1]

    g = f.createVariable(
        name,
        np.float32,
        dims,
        zlib=True,
        complevel=1,
        shuffle=True,
        fletcher32=True,
        fill_value=np.nan,
    )

    if value.ndim > 1:
        g[:] = value.transpose()
    else:
        g[:] = value


def Efield(outdir: Path, E: xarray.Dataset):
    """
    write Efield to disk
    """

    if Dataset is None:
        raise ImportError("pip install netcdf4")

    with Dataset(outdir / "simsize.nc", "w") as f:
        g = f.createVariable("llon", np.int32)
        g[:] = E.mlon.size
        g = f.createVariable("llat", np.int32)
        g[:] = E.mlat.size

    with Dataset(outdir / "simgrid.nc", "w") as f:
        f.createDimension("mlon", E.mlon.size)
        f.createDimension("mlat", E.mlat.size)
        _write_var(f, "mlon", E.mlon)
        _write_var(f, "mlat", E.mlat)

    for t in E.time:
        time: datetime = to_datetime(t)
        fn = outdir / (datetime2ymd_hourdec(time) + ".nc")

        # FOR EACH FRAME WRITE A BC TYPE AND THEN OUTPUT BACKGROUND AND BCs
        with Dataset(fn, "w") as f:
            f.createDimension("mlon", E.mlon.size)
            f.createDimension("mlat", E.mlat.size)

            g = f.createVariable("flagdirich", np.int32)
            g[:] = E["flagdirich"].loc[time]

            f.createDimension("ymd", 3)
            g = f.createVariable("ymd", np.int32, "ymd")
            g[:] = [time.year, time.month, time.day]

            g = f.createVariable("UTsec", np.float32)
            g[:] = time.hour * 3600 + time.minute * 60 + time.second + time.microsecond / 1e6

            for k in {"Exit", "Eyit", "Vminx1it", "Vmaxx1it"}:
                _write_var(f, k, E[k].loc[time])

            for k in {"Vminx2ist", "Vmaxx2ist"}:
                _write_var(f, k, E[k].loc[time])

            for k in {"Vminx3ist", "Vmaxx3ist"}:
                _write_var(f, k, E[k].loc[time])


def precip(outdir: Path, P: xarray.Dataset):
    """
    write precipitation to disk
    """

    if Dataset is None:
        raise ImportError("pip install netcdf4")

    with Dataset(outdir / "simsize.nc", "w") as f:
        g = f.createVariable("llon", np.int32)
        g[:] = P.mlon.size
        g = f.createVariable("llat", np.int32)
        g[:] = P.mlat.size

    with Dataset(outdir / "simgrid.nc", "w") as f:
        f.createDimension("mlon", P.mlon.size)
        f.createDimension("mlat", P.mlat.size)
        _write_var(f, "mlon", P.mlon)
        _write_var(f, "mlat", P.mlat)

    for t in P.time:
        time: datetime = to_datetime(t)
        fn = outdir / (datetime2ymd_hourdec(time) + ".nc")

        with Dataset(fn, "w") as f:
            f.createDimension("mlon", P.mlon.size)
            f.createDimension("mlat", P.mlat.size)

            for k in {"Q", "E0"}:
                _write_var(f, f"{k}p", P[k].loc[time])



"""
setup a new simulation
"""

from __future__ import annotations
import logging
import importlib
import argparse
from pathlib import Path
import typing as T
import shutil

from .config import read_nml
from . import grid
from .plasma import equilibrium_state, equilibrium_resample
from .efield import Efield_BCs
from .particles import particles_BCs
from . import namelist
from . import write

__all__ = ["setup", "config"]


def config(params: dict[str, T.Any], out_dir: Path):
    """
    top-level API to create a new simulation.
    This is meant to be a front-end to the model.setup() function, creating
    files expected by model.setup(), especially config.nml

    PySat and other Python programs will normally use this function, perhaps via a shim
    to their internal data structures.

    The required namelists are written one-by-one for best modularity.
    The order of namelists does not matter, nor does the order of variables in a namelist.

    Parameters
    ----------

    params: dict
        location, time, and extents thereof
    """

    nml_file = out_dir / "inputs/config.nml"
    nml_file.parent.mkdir(parents=True, exist_ok=True)

    logging.info(f"Creating Gemini3D configuration file under {nml_file}")

    file_format = params.get("file_format", "h5")

    # %% base
    t0 = params["time"][0]
    tend = params["time"][-1]

    base = {
        "ymd": [t0.year, t0.month, t0.day],
        "UTsec0": t0.hour * 3600 + t0.minute * 60 + t0.second + t0.microsecond / 1e6,
        "tdur": (tend - t0).total_seconds(),
        "dtout": params["dtout"],
        "activ": [params["f107a"], params["f107"], params["Ap"]],
        "tcfl": params.get("tcfl", 0.9),
        "Teinf": params.get("Teinf", 1500.0),
    }
    namelist.write(nml_file, namelist="base", data=base)

    # %% flags
    flags = {"flagoutput": params.get("flagoutput", 1), "potsolve": params.get("potsolve", 1)}
    namelist.write(nml_file, "flags", flags)

    # %% files
    files = {
        "indat_size": f"inputs/simsize.{file_format}",
        "indat_grid": f"inputs/simgrid.{file_format}",
        "indat_file": f"inputs/initial_conditions.{file_format}",
    }
    namelist.write(nml_file, "files", files)

    # %% setup
    setup = {
        "glat": params["glat"],
        "glon": params["glon"],
        "xdist": params["x2dist"],
        "ydist": params["x3dist"],
        "alt_min": params["alt_min"],
        "alt_max": params["alt_max"],
        "lxp": params["lx2"],
        "lyp": params["lx3"],
        "Bincl": params["Bincl"],
        "nmf": params["Nmf"],
        "nme": params["Nme"],
    }

    for k in "eq_dir":
        if k in params:
            setup[k] = params[k]

    namelist.write(nml_file, "setup", setup)


def setup(path: Path | dict[str, T.Any], out_dir: Path):
    """
    top-level function to create a new simulation FROM A FILE config.nml

    Parameters
    ----------

    path: pathlib.Path
        path (directory or full path) to config.nml
    out_dir: pathlib.Path
        directory to write simulation artifacts to
    """

    # %% read config.nml
    if isinstance(path, dict):
        cfg = path
    elif isinstance(path, (str, Path)):
        cfg = read_nml(path)
    else:
        raise TypeError("expected Path to config.nml or dict with parameters")

    if not cfg:
        raise FileNotFoundError(f"no configuration found for {out_dir}")

    cfg["out_dir"] = Path(out_dir).expanduser().resolve()

    for k in ("indat_size", "indat_grid", "indat_file"):
        cfg[k] = cfg["out_dir"] / cfg[k]

    # FIXME: should use is_absolute() ?
    for k in ("eq_dir", "eqzip", "E0dir", "precdir"):
        if cfg.get(k):
            cfg[k] = (cfg["out_dir"] / cfg[k]).resolve()

    # %% copy input config.nml to output dir
    input_dir = cfg["out_dir"] / "inputs"
    input_dir.mkdir(parents=True, exist_ok=True)
    shutil.copy2(cfg["nml"], input_dir)

    # %% is this equilibrium or interpolated simulation
    if "eq_dir" in cfg:
        interp(cfg)
    else:
        equilibrium(cfg)


def equilibrium(p: dict[str, T.Any]):
    # %% GRID GENERATION

    xg = grid.cart3d(p)

    write.grid(p, xg)

    # %% Equilibrium input generation
    dat = equilibrium_state(p, xg)

    write.state(p["indat_file"], dat)


def interp(cfg: dict[str, T.Any]) -> None:

    xg = grid.cart3d(cfg)

    equilibrium_resample(cfg, xg)

    postprocess(cfg, xg)


def postprocess(cfg: dict[str, T.Any], xg: dict[str, T.Any]) -> None:
    """
    defaults to applying config.nml defined E-field and/or precipitation

    However, the user can also apply their own functions in config.nml
    &setup setup_functions
    """

    if "setup_functions" in cfg:
        # assume file to import from is in cwd or Python path
        funcs = (
            [cfg["setup_functions"]]
            if isinstance(cfg["setup_functions"], str)
            else cfg["setup_functions"]
        )

        for name in funcs:
            mod_name = ".".join(name.split(".")[:-1])
            func_name = name.split(".")[-1]
            if not mod_name:
                # file with single function of same name e.g. perturb.py with function perturb()
                mod_name = func_name
            mod = importlib.import_module(mod_name)
            getattr(mod, func_name)(cfg, xg)

        return

    # %% potential boundary conditions
    if "E0dir" in cfg:
        Efield_BCs(cfg, xg)

    # %% aurora
    if "precdir" in cfg:
        particles_BCs(cfg, xg)


def cli():
    p = argparse.ArgumentParser()
    p.add_argument("config_file", help="path to config*.nml file")
    p.add_argument("out_dir", help="simulation output directory")
    P = p.parse_args()

    setup(P.config_file, P.out_dir)


if __name__ == "__main__":
    cli()



"""
functions for finding files
"""

from datetime import datetime, timedelta
from pathlib import Path
import numpy as np

FILE_FORMATS = [".h5", ".nc", ".dat"]


def config(path: Path) -> Path:
    """ given a path or config filename, return the full path to config file """

    return find_stem(path, "config", "nml")


def simsize(path: Path, suffix: str = None) -> Path:
    """ gets path to simsize file """

    return find_stem(path, "simsize", suffix)


def frame(simdir: Path, time: datetime, file_format: str = None) -> Path:
    """
    the frame filenames can have different file formats
    """

    simdir = Path(simdir).expanduser()

    stem = (
        time.strftime("%Y%m%d")
        + f"_{time.hour*3600 + time.minute*60 + time.second:05d}."
        + f"{time.microsecond:06d}"
    )

    suffixes = [f".{file_format}"] if file_format else FILE_FORMATS

    for ext in suffixes:
        fn = simdir / (stem + ext)
        if fn.is_file():
            return fn

    # %% WORKAROUND for real32 file ticks. This will be removed when datetime-fortran is implemented
    MAX_OFFSET = timedelta(seconds=1)  # 10 ms precision, allow extra accumulated tolerance
    pat = time.strftime("%Y%m%d") + "_*"
    for ext in suffixes:
        file_times = []
        files = list(simdir.glob(pat + ext))
        for fn in files:
            file_times.append(
                datetime.strptime(fn.name[:8], "%Y%m%d") + timedelta(seconds=float(fn.name[9:21]))
            )

        if file_times:
            afile_times = np.array(file_times)
            i = abs(afile_times - time).argmin()  # type: ignore

            if abs(afile_times[i] - time) <= MAX_OFFSET:
                return files[i]

    return None


def grid(path: Path) -> Path:
    """given a path or filename, return the full path to simgrid file
    we don't override FILE_FORMATS to allow outputs from a prior sim in a different
    file format to be used in this sim.
    """

    return find_stem(path, "simgrid")


def find_stem(path: Path, stem: str, suffix: str = None) -> Path:
    """find file containing stem """

    path = Path(path).expanduser()

    if path.is_file():
        if stem in path.stem:
            return path
        else:
            return find_stem(path.parent, stem, path.suffix)

    if suffix:
        if isinstance(suffix, str):
            if not suffix.startswith("."):
                suffix = "." + suffix
            suffixes = [suffix]
        else:
            suffixes = suffix
    else:
        suffixes = FILE_FORMATS

    if path.is_dir():
        for p in (path, path / "inputs"):
            for suff in suffixes:
                f = p / (stem + suff)
                if f.is_file():
                    return f

    return None


def inputs(direc: Path, input_dir: Path = None) -> Path:
    """
    find input parameter directory

    direc: pathlib.Path
        top level simulation dir
    input_dir: pathlib.Path
        relative to top level, the parameter dir. E.g. inputs/precip, input/Efield
    """

    direc = Path(direc).expanduser()
    if input_dir:
        input_path = Path(input_dir).expanduser()
        if not input_path.is_absolute():
            input_path = direc / input_path
    else:
        input_path = direc

    if not input_path.is_dir():
        input_path = None

    return input_path



"""
functions related to running gemini.bin.
Either locally (laptop or interactive HPC) or creating an HPC batch script based on template.

Builds gemini.bin if not found.
"""

from __future__ import annotations
import typing as T
import os
import logging
import subprocess
import shutil
from pathlib import Path
import numpy as np
import psutil

from . import find
from . import mpi
from . import cmake
from .hpc import hpc_batch_detect, hpc_batch_create
from . import model
from . import write
from . import read
from .utils import git_meta


Pathlike = T.Union[str, Path]


def runner(pr: dict[str, T.Any]) -> None:

    out_dir = check_outdir(pr["out_dir"])

    config_file = find.config(pr["config_file"])
    # load configuration to know what directories to check
    p = read.config(config_file)
    if not p:
        raise FileNotFoundError(f"{config_file} does not appear to contain config.nml")

    # we don't want to overwrite an expensive simulation output
    if find.frame(out_dir, p["time"][0], p.get("out_format")):
        raise FileExistsError(
            f"a fresh simulation should not have data in output directory: {out_dir}"
        )

    # %% setup grid and/or initial ionosphere state if needed
    for k in ("indat_size", "indat_grid", "indat_file"):
        f = out_dir / p[k].expanduser()
        if pr.get("force") or not f.is_file():
            model.setup(p["nml"], out_dir)

    # %% estimate simulation RAM use on root MPI node
    ram_use_bytes = memory_estimate(out_dir)

    # %% setup Efield if needed
    if "E0dir" in p:
        E0dir = out_dir / p["E0dir"]
        if not E0dir.is_dir():
            model.setup(p["nml"], out_dir)

    # %% setup precip if needed
    if "precdir" in p:
        precdir = out_dir / p["precdir"]
        if not precdir.is_dir():
            model.setup(p["nml"], out_dir)

    # build checks
    gemexe = get_gemini_exe(pr.get("gemexe"))
    logging.info(f"gemini executable: {gemexe}")

    mpiexec = check_mpiexec(pr.get("mpiexec"), gemexe)
    if mpiexec:
        logging.info(f"mpiexec: {mpiexec}")
        Nmpi = mpi.count(out_dir / p["indat_size"], pr.get("cpu_count"))
        mpi_cmd = [mpiexec, "-n", str(Nmpi)]
    else:
        mpi_cmd = []

    cmd = mpi_cmd + [str(gemexe), str(out_dir)]

    if pr.get("out_format"):
        cmd += ["-out_format", pr["out_format"]]

    # %% attempt dry run, but don't fail in case intended for HPC
    logging.info("Gemini dry run command:")
    logging.info(" ".join(cmd))
    proc = subprocess.run(cmd + ["-dryrun"])

    if proc.returncode != 0:
        raise RuntimeError("Gemini dry run failed.")

    if pr.get("dryrun"):
        return None

    write.meta(out_dir / "setup_meta.nml", git_meta(gemexe.parent), "setup_gemini")

    batcher = hpc_batch_detect()
    if batcher:
        job_file = hpc_batch_create(batcher, out_dir, cmd)
        print("Please examine batch file", job_file, "and when ready submit the job as usual.")
    else:
        avail_memory = psutil.virtual_memory().available
        if avail_memory < 2 * ram_use_bytes:
            logging.warning(
                f"""
Computer RAM available: {avail_memory/1e9:.1} GB but simulation needs {ram_use_bytes/1e9:.1}
Gemini3D may run out of RAM on this computer, which may make the run exceedingly slow or fail.
"""
            )
        print("\nBEGIN Gemini run with command:")
        print(" ".join(cmd), "\n")
        ret = subprocess.run(cmd).returncode
        if ret != 0:
            raise RuntimeError("Gemini run failed")


def memory_estimate(path: Path) -> int:
    """
    Estimate how must RAM gemini.bin will need.
    The current Gemini MPI architecture assumes the root node will use the most RAM,
    so that is the fundamental constraint.
    This neglects size of executable and library footprint,
    which would be miniscule in simulations larger than 1 GB where we
    are concerned about RAM limits.

    Parameters
    ----------

    path: pathlib.Path
        path to simgrid.*

    Returns
    -------

    memory_used: int
        estimated RAM usage (bytes)

    """

    SIZE = 8  # number of bytes for each element: real64 => 8 bytes
    PAD = 2  # factor to assume needed over variable itself (computations, work variables, ...)

    gs = read.grid(path, shape=True)

    grid_size = 0

    for k, v in gs.items():
        if k == "lxs" or not isinstance(v, (tuple, list, np.ndarray)):
            continue
        grid_size += np.prod(v)  # type: ignore

    LSP = 7
    x1 = gs["x1"][0]
    x2 = gs["x2"][0]
    x3 = gs["x3"][0]

    Ns = LSP * x1 * x2 * x3
    Ts = Ns

    memory_used = SIZE * (grid_size + (Ns + Ts) * PAD)

    return memory_used


def check_compiler():

    fc = os.environ.get("FC")
    fc = shutil.which(fc) if fc else shutil.which("gfortran")
    if not fc:
        raise EnvironmentError("Cannot find Fortran compiler e.g. Gfortran")


def check_mpiexec(mpiexec: Pathlike, gemexe: Pathlike) -> str:
    """check if specified mpiexec exists on this system
    If not, fall back to not using MPI (slow, but still works)."""

    if not mpiexec:
        mpiexec = "mpiexec"
    mpi_root = os.environ.get("MPI_ROOT")
    if mpi_root:
        mpi_root += "/bin"

    mpiexec = shutil.which(mpiexec, path=str(mpi_root))
    if not mpiexec:
        return None

    ret = subprocess.run([mpiexec, "-help"], stdout=subprocess.PIPE, text=True, timeout=5)
    if ret.returncode != 0:
        return None
    # %% check that compiler and MPIexec compatible
    if os.name != "nt":
        return mpiexec

    mpi_msg = ret.stdout.strip()
    ret = subprocess.run([str(gemexe), "-compiler"], stdout=subprocess.PIPE, text=True, timeout=5)
    if ret.returncode != 0:
        raise EnvironmentError(f"{gemexe} not executable")

    if "GNU" in ret.stdout.strip() and "Intel(R) MPI Library" in mpi_msg:
        mpiexec = None
        logging.error("Not using MPIexec since MinGW is not compatible with Intel MPI")

    return mpiexec


def get_gemini_exe(gemexe: Path = None) -> Path:
    """
    find and check that Gemini exectuable can run on this system
    download and build Gemini3D if needed
    """

    if not gemexe:  # allow for default dict empty
        gemexe = Path("gemini.bin")
    gemexe = cmake.build_gemini3d(gemexe)

    # %% ensure gemini.bin is runnable
    ret = subprocess.run([str(gemexe)], stdout=subprocess.DEVNULL, timeout=15)
    if ret.returncode != 0:
        raise RuntimeError(
            f"\n{gemexe} was not runnable on your platform. Try recompiling on this computer type."
            "E.g. different HPC nodes may not have the CPU feature sets."
        )

    return gemexe


def check_outdir(out_dir: Pathlike) -> Path:

    out_dir = Path(out_dir).expanduser().resolve()
    if out_dir.is_file():
        raise NotADirectoryError(f"please specify output DIRECTORY, you specified {out_dir}")
    if not out_dir.is_dir():
        out_dir.mkdir(parents=True, exist_ok=True)

    return out_dir



import logging
import argparse
from pathlib import Path
import time

from .job import runner


def cli():
    p = argparse.ArgumentParser()
    p.add_argument("config_file", help="path to config*.nml file")
    p.add_argument("out_dir", help="simulation output directory")
    p.add_argument("-dryrun", help="quick check if sim is ok", action="store_true")
    p.add_argument("-mpiexec", help="path to desired mpiexec executable")
    p.add_argument("-gemexe", help="path to desired gemini.bin")
    p.add_argument("-n", "--cpu", help="number of CPU cores", type=int, default=0)
    p.add_argument("-f", "--force", help="force regeneration of simulation", action="store_true")
    p.add_argument(
        "-out_format", help="override Fortran output file format", choices=["h5", "nc", "raw"]
    )
    p.add_argument("-v", "--verbose", action="store_true")
    P = p.parse_args()

    level = logging.INFO if P.verbose else None
    logging.basicConfig(format="%(message)s", level=level)

    params = {
        "config_file": Path(P.config_file).expanduser(),
        "out_dir": Path(P.out_dir).expanduser().resolve(),
        "mpiexec": P.mpiexec,
        "gemexe": P.gemexe,
        "force": P.force,
        "out_format": P.out_format,
        "cpu_count": P.cpu,
        "dryrun": P.dryrun,
    }

    tic = time.monotonic()
    runner(params)
    print(f"job.py ran in {time.monotonic() - tic:.3f} seconds.")


if __name__ == "__main__":
    cli()



"""
HDF5 file writing
"""

from __future__ import annotations
import xarray
import typing as T
import numpy as np
from pathlib import Path
from datetime import datetime
import logging

from .. import LSP
from ..utils import datetime2ymd_hourdec, to_datetime

try:
    import h5py
except (ImportError, AttributeError):
    # must be ImportError not ModuleNotFoundError for botched HDF5 linkage
    h5py = None


def state(fn: Path, dat: xarray.Dataset):
    """
    write STATE VARIABLE initial conditions

    NOTE THAT WE don't write ANY OF THE ELECTRODYNAMIC
    VARIABLES SINCE THEY ARE NOT NEEDED TO START THINGS
    UP IN THE FORTRAN CODE.

    INPUT ARRAYS SHOULD BE TRIMMED TO THE CORRECT SIZE
    I.E. THEY SHOULD NOT INCLUDE GHOST CELLS
    """

    if h5py is None:
        raise ImportError("pip install h5py")

    time = dat.time

    logging.info(f"state: {fn} {time}")

    with h5py.File(fn, "w") as f:
        f["/time/ymd"] = [time.year, time.month, time.day]
        f["/time/UTsec"] = np.float32(
            time.hour * 3600 + time.minute * 60 + time.second + time.microsecond / 1e6
        )

        for k in {"ns", "vs1", "Ts"}:
            if k in dat.data_vars:
                _write_var(f, f"/{k}all", dat[k])

        if "Phitop" in dat.data_vars:
            _write_var(f, "/Phiall", dat["Phitop"])


def _write_var(f, name: str, A: xarray.DataArray):
    """
    NOTE: The .transpose() reverses the dimension order.
    The HDF Group never implemented the intended H5T_array_create(..., perm)
    and it's deprecated.

    Fortran, including the HDF Group Fortran interfaces and h5fortran as well as
    Matlab read/write HDF5 in Fortran order. h5py read/write HDF5 in C order so we
    need the .transpose() for h5py
    """

    p4s = ("species", "x3", "x2", "x1")

    if A.ndim == 4:
        A = A.transpose(*p4s)
    elif A.ndim == 2:
        A = A.transpose()
    elif A.ndim == 1:
        pass
    else:
        raise ValueError(
            f"write_hdf5: unexpected number of dimensions {A.ndim}. Please raise a GitHub Issue."
        )

    f.create_dataset(
        name,
        data=A,
        dtype=np.float32,
        compression="gzip",
        compression_opts=1,
        shuffle=True,
        fletcher32=True,
    )


def data(outfn: Path, dat: xarray.Dataset):
    """
    write simulation data
    e.g. for converting a file format from a simulation
    """

    if h5py is None:
        raise ImportError("pip install h5py")

    lxs = dat.shape

    with h5py.File(outfn, "w") as h:
        for k in {"ns", "vs1", "Ts"}:
            if k not in dat:
                continue

            h.create_dataset(
                k,
                data=dat[k].astype(np.float32),
                chunks=(1, *lxs[1:], LSP),
                compression="gzip",
                compression_opts=1,
            )

        for k in {"ne", "v1", "Ti", "Te", "J1", "J2", "J3", "v2", "v3"}:
            if k not in dat:
                continue

            h.create_dataset(
                k,
                data=dat[k].astype(np.float32),
                chunks=(1, *lxs[1:]),
                compression="gzip",
                compression_opts=1,
            )

        if "Phitop" in dat:
            h.create_dataset(
                "Phiall",
                data=dat["Phitop"].astype(np.float32),
                compression="gzip",
                compression_opts=1,
            )


def grid(size_fn: Path, grid_fn: Path, xg: dict[str, T.Any]):
    """writes grid to disk

    Parameters
    ----------

    size_fn: pathlib.Path
        file to write
    grid_fn: pathlib.Path
        file to write
    xg: dict
        grid values

    NOTE: The .transpose() reverses the dimension order.
    The HDF Group never implemented the intended H5T_array_create(..., perm)
    and it's deprecated.
    Fortran, including the HDF Group Fortran interfaces and h5fortran as well as
    Matlab read/write HDF5 in Fortran order. h5py read/write HDF5 in C order so we
    need the .transpose() for h5py
    """

    if h5py is None:
        raise ImportError("pip install h5py")

    if "lx" not in xg:
        xg["lx"] = np.array([xg["x1"].shape, xg["x2"].shape, xg["x3"].shape])

    logging.info(f"write_grid: {size_fn}")
    with h5py.File(size_fn, "w") as h:
        h["/lx"] = xg["lx"]

    logging.info(f"write_grid: {grid_fn}")
    with h5py.File(grid_fn, "w") as h:
        for i in (1, 2, 3):
            for k in (
                f"x{i}",
                f"x{i}i",
                f"dx{i}b",
                f"dx{i}h",
                f"h{i}",
                f"h{i}x1i",
                f"h{i}x2i",
                f"h{i}x3i",
                f"gx{i}",
                f"e{i}",
            ):
                if k not in xg:
                    logging.info(f"SKIP: {k}")
                    continue

                if xg[k].ndim >= 2:
                    h.create_dataset(
                        f"/{k}",
                        data=xg[k].transpose(),
                        dtype=np.float32,
                        compression="gzip",
                        compression_opts=1,
                        shuffle=True,
                        fletcher32=True,
                    )
                else:
                    h[f"/{k}"] = xg[k].astype(np.float32)

        for k in (
            "alt",
            "glat",
            "glon",
            "Bmag",
            "I",
            "nullpts",
            "er",
            "etheta",
            "ephi",
            "r",
            "theta",
            "phi",
            "x",
            "y",
            "z",
        ):
            if k not in xg:
                logging.info(f"SKIP: {k}")
                continue

            if xg[k].ndim >= 2:
                h.create_dataset(
                    f"/{k}",
                    data=xg[k].transpose(),
                    dtype=np.float32,
                    compression="gzip",
                    compression_opts=1,
                    shuffle=True,
                    fletcher32=True,
                )
            else:
                h[f"/{k}"] = xg[k].astype(np.float32)


def Efield(outdir: Path, E: xarray.Dataset):
    """
    write Efield to disk
    """

    if h5py is None:
        raise ImportError("pip install h5py")

    with h5py.File(outdir / "simsize.h5", "w") as f:
        f["/llon"] = E.mlon.size
        f["/llat"] = E.mlat.size

    with h5py.File(outdir / "simgrid.h5", "w") as f:
        f["/mlon"] = E.mlon.astype(np.float32)
        f["/mlat"] = E.mlat.astype(np.float32)

    for t in E.time:
        time: datetime = to_datetime(t)
        fn = outdir / (datetime2ymd_hourdec(time) + ".h5")

        # FOR EACH FRAME WRITE A BC TYPE AND THEN OUTPUT BACKGROUND AND BCs
        with h5py.File(fn, "w") as f:
            f["/flagdirich"] = E["flagdirich"].loc[time].astype(np.int32)
            f["/time/ymd"] = [time.year, time.month, time.day]
            f["/time/UTsec"] = (
                time.hour * 3600 + time.minute * 60 + time.second + time.microsecond / 1e6
            )

            for k in {"Exit", "Eyit", "Vminx1it", "Vmaxx1it"}:
                f.create_dataset(
                    f"/{k}",
                    data=E[k].loc[time].transpose(),
                    dtype=np.float32,
                    compression="gzip",
                    compression_opts=1,
                    shuffle=True,
                    fletcher32=True,
                )
            for k in {"Vminx2ist", "Vmaxx2ist", "Vminx3ist", "Vmaxx3ist"}:
                f[f"/{k}"] = E[k].loc[time].astype(np.float32)


def precip(outdir: Path, P: xarray.Dataset):

    if h5py is None:
        raise ImportError("pip install h5py")

    with h5py.File(outdir / "simsize.h5", "w") as f:
        f["/llon"] = P.mlon.size
        f["/llat"] = P.mlat.size

    with h5py.File(outdir / "simgrid.h5", "w") as f:
        f["/mlon"] = P.mlon.astype(np.float32)
        f["/mlat"] = P.mlat.astype(np.float32)

    for t in P.time:
        time: datetime = to_datetime(t)
        fn = outdir / (datetime2ymd_hourdec(time) + ".h5")

        with h5py.File(fn, "w") as f:
            for k in {"Q", "E0"}:
                f.create_dataset(
                    f"/{k}p",
                    data=P[k].loc[time].transpose(),
                    dtype=np.float32,
                    compression="gzip",
                    compression_opts=1,
                    shuffle=True,
                    fletcher32=True,
                )



from __future__ import annotations
import subprocess
import os
import re
import shutil
from pathlib import Path
import xarray
import numpy as np
from datetime import datetime
import typing as T
import logging

try:
    import psutil
except ImportError:
    psutil = None
    # pip install psutil will improve CPU utilization.

Pathlike = T.Union[str, Path]

__all__ = ["to_datetime", "git_meta", "get_cpu_count", "datetime2ymd_hourdec"]


def to_datetime(times: xarray.DataArray | np.datetime64 | datetime) -> datetime:
    """

    Parameters
    ----------
    atimes : xarray time

    Returns
    -------
    times : list[datetime.datetime]
    """

    if isinstance(times, datetime):
        time = times
    elif isinstance(times, xarray.DataArray):
        time = times.values.squeeze()[()]  # numpy.datetime64
    elif isinstance(times, np.datetime64):
        time = times.squeeze()[()]
    else:
        raise TypeError("expected datetime-like value")

    if isinstance(time, np.datetime64):
        time = time.astype("datetime64[us]").astype(datetime)

    return time


def git_meta(path: Path = None) -> dict[str, str]:
    """
    provide metadata about a Git repo in a dictionary

    Dev note: use subprocess.run to avoid crashing program when Git meta is missing or broken (shallow clone)

    empty init in case can't read Git info
    this avoids needless if statements in consumers
    """

    git = shutil.which("git")
    meta = {
        "git_version": None,
        "remote": None,
        "branch": None,
        "commit": None,
        "porcelain": "false",
    }
    if not git:
        return meta

    if not path:
        if __file__:
            path = Path(__file__).resolve().parent
        else:
            return meta

    ret = subprocess.run([git, "-C", str(path), "--version"], stdout=subprocess.PIPE, text=True)
    if ret.returncode != 0:
        logging.error("Git was not available or is too old")
        return meta

    meta["git_version"] = ret.stdout.strip()

    ret = subprocess.run([git, "-C", str(path), "rev-parse"])
    if ret.returncode != 0:
        logging.error(f"{path} is not a Git repo.")
        return meta

    ret = subprocess.run(
        [git, "-C", str(path), "rev-parse", "--abbrev-ref", "HEAD"],
        stdout=subprocess.PIPE,
        text=True,
    )
    if ret.returncode != 0:
        logging.error(f"{path} coul not determine Git branch")
        return meta
    meta["branch"] = ret.stdout.strip()

    ret = subprocess.run([git, "-C", str(path), "remote", "-v"], stdout=subprocess.PIPE, text=True)
    if ret.returncode != 0:
        logging.error(f"{path} could not determine Git remote")
        return meta

    pat = re.compile(r"^origin\s+(.*)\s+\(fetch\)")
    mat = pat.match(ret.stdout.strip())
    if mat:
        meta["remote"] = mat.group(1)

    ret = subprocess.run(
        [git, "-C", str(path), "rev-parse", "--short", "HEAD"], stdout=subprocess.PIPE, text=True
    )
    if ret.returncode != 0:
        logging.error(f"{path} could not determine Git commit")
        return meta
    meta["commit"] = ret.stdout.strip()

    ret = subprocess.run(
        [git, "-C", str(path), "status", "--porcelain"], stdout=subprocess.PIPE, text=True
    )
    if ret.returncode != 0:
        logging.error(f"{path} could not determine Git status")
    msg = ret.stdout.strip()
    meta["porcelain"] = "false" if msg else "true"

    return meta


def get_cpu_count() -> int:
    """get a physical CPU count

    Note: len(os.sched_getaffinity(0)) and multiprocessing.cpu_count don't help either
    PSUtil is the most reliable, so we strongly recommend it.

    Returns
    -------
    count: int
        detect number of physical CPU
    """

    extradiv = 1
    max_cpu = None
    # without psutil, hyperthreaded CPU may overestimate physical count by factor of 2 (or more)
    if psutil is not None:
        max_cpu = psutil.cpu_count(logical=False)
        if max_cpu is None:
            max_cpu = psutil.cpu_count()
            extradiv = 2
    if max_cpu is None:
        max_cpu = os.cpu_count()
        if max_cpu is not None:
            extradiv = 2
        else:
            max_cpu = 1

    return max_cpu // extradiv


def datetime2ymd_hourdec(dt: datetime) -> str:
    """
    convert datetime to ymd_hourdec string for filename stem
    """

    dt = to_datetime(dt)

    assert isinstance(dt, datetime), "expect scalar datetime"

    return (
        dt.strftime("%Y%m%d")
        + f"_{dt.hour*3600 + dt.minute*60 + dt.second + dt.microsecond/1e6:12.6f}"
    )



from __future__ import annotations
import typing as T
import re
import math
from pathlib import Path
from datetime import datetime, timedelta

from . import find
from . import namelist

NaN = math.nan


def datetime_range(start: datetime, stop: datetime, step: timedelta) -> list[datetime]:

    """
    Generate range of datetime over a closed interval.
    pandas.date_range also defaults to a closed interval.
    That means that the start AND stop time are included.

    Parameters
    ----------
    start : datetime
        start time
    stop : datetime
        stop time
    step : timedelta
        time step

    Returns
    -------
    times : list of datetime
        times requested
    """

    return [start + i * step for i in range((stop - start) // step + 1)]


def read_nml(fn: Path) -> dict[str, T.Any]:
    """parse .nml file
    for now we don't use the f90nml package, though maybe we will in the future.
    Just trying to keep Python prereqs reduced for this simple parsing.
    """

    params: dict[str, T.Any] = {}

    fn = find.config(fn)
    if not fn:
        return params

    params["nml"] = fn
    for n in ("base", "files", "flags", "setup"):
        params.update(parse_namelist(fn, n))

    if namelist_exists(fn, "neutral_perturb"):
        params.update(parse_namelist(fn, "neutral_perturb"))
    if namelist_exists(fn, "precip"):
        params.update(parse_namelist(fn, "precip"))
    if namelist_exists(fn, "efield"):
        params.update(parse_namelist(fn, "efield"))
    if namelist_exists(fn, "glow"):
        params.update(parse_namelist(fn, "glow"))

    return params


def namelist_exists(fn: Path, nml: str) -> bool:
    """ determines if a namelist exists in a file """

    pat = re.compile(r"^\s*&(" + nml + ")$")

    with fn.open("r") as f:
        for line in f:
            if pat.match(line) is not None:
                return True

    return False


def parse_namelist(file: Path, nml: str) -> dict[str, T.Any]:
    """
    this is Gemini-specific
    don't resolve absolute paths here because that assumes same machine
    """

    r = namelist.read(file, nml)

    P: dict[str, T.Any] = {}

    if nml == "base":
        t0 = datetime(int(r["ymd"][0]), int(r["ymd"][1]), int(r["ymd"][2])) + timedelta(
            seconds=float(r["UTsec0"])
        )
        P["tdur"] = timedelta(seconds=float(r["tdur"]))
        P["dtout"] = timedelta(seconds=float(r["dtout"]))
        P["time"] = datetime_range(t0, t0 + P["tdur"], P["dtout"])

        P["f107a"] = float(r["activ"][0])
        P["f107"] = float(r["activ"][1])
        P["Ap"] = float(r["activ"][2])
        P["tcfl"] = float(r["tcfl"])
        P["Teinf"] = float(r["Teinf"])
    elif nml == "flags":
        for k in r:
            P[k] = int(r[k])
    elif nml == "files":
        for k in ("indat_file", "indat_grid", "indat_size"):
            P[k] = Path(r[k]).expanduser()

        P["file_format"] = r.get("file_format", P["indat_size"].suffix[1:])
        # defaults to type of input

        if "realbits" in r:
            P["realbits"] = int(r["realbits"])
        else:
            if P["file_format"] in ("raw", "dat"):
                P["realbits"] = 64
            else:
                P["realbits"] = 32
    elif nml == "setup":
        for k in r:
            if k in {"lxp", "lyp"}:
                P[k] = int(r[k])
            elif k == "eqdir":  # old .nml
                P["eq_dir"] = Path(r[k]).expanduser()
            elif k == "eq_dir":
                P["eq_dir"] = Path(r[k]).expanduser()
            elif k == "setup_functions":
                P["setup_functions"] = [r[k]] if isinstance(r[k], str) else r[k]
            else:
                P[k] = r[k]
    elif nml == "neutral_perturb":
        P["interptype"] = int(r["interptype"])
        P["sourcedir"] = Path(r["source_dir"]).expanduser()

        for k in ("sourcemlat", "sourcemlon", "dtneu", "dxn", "drhon", "dzn"):
            try:
                P[k] = float(r[k])
            except KeyError:
                P[k] = NaN
    elif nml == "precip":
        P["dtprec"] = timedelta(seconds=float(r["dtprec"]))
        P["precdir"] = Path(r["prec_dir"]).expanduser()
    elif nml == "efield":
        P["dtE0"] = timedelta(seconds=float(r["dtE0"]))
        P["E0dir"] = Path(r["E0_dir"]).expanduser()
    elif nml == "glow":
        P["dtglow"] = timedelta(seconds=float(r["dtglow"]))
        P["dtglowout"] = float(r["dtglowout"])

    if not P:
        raise ValueError(f"Not sure how to parse NML namelist {nml}")

    return P


def read_ini(fn: Path) -> dict[str, T.Any]:
    """ parse .ini file (legacy) """

    P: dict[str, T.Any] = {}

    fn = find.config(fn)

    if not fn:
        return P

    with fn.open("r") as f:
        date = list(map(int, f.readline().split()[0].split(",")))[::-1]
        sec = float(f.readline().split()[0])
        t0 = datetime(date[0], date[1], date[2]) + timedelta(seconds=sec)
        P["tdur"] = timedelta(seconds=float(f.readline().split()[0]))
        P["dtout"] = timedelta(seconds=float(f.readline().split()[0]))
        P["time"] = datetime_range(t0, t0 + P["tdur"], P["dtout"])

        P["f107a"], P["f107"], P["Ap"] = map(float, f.readline().split()[0].split(","))

        P["tcfl"] = float(f.readline().split()[0])
        P["Teinf"] = float(f.readline().split()[0])

        P["potsolve"] = int(f.readline().split()[0])
        P["flagperiodic"] = int(f.readline().split()[0])
        P["flagoutput"] = int(f.readline().split()[0])
        P["flagcap"] = int(f.readline().split()[0])

        for k in ("indat_size", "indat_grid", "indat_file"):
            P[k] = Path(f.readline().strip().replace("'", "").replace('"', "")).expanduser()

    return P



import pytest
from datetime import datetime, timedelta
import importlib.resources

import gemini3d.config as config
import gemini3d.read as read
import gemini3d.model as model


def test_model_config(tmp_path):
    P = {
        "time": [
            datetime(2020, 1, 1, 5, 1, 3),
            datetime(2020, 1, 1, 5, 1, 4),
            datetime(2020, 1, 1, 5, 1, 5),
        ],
        "dtout": 1.0,
        "f107a": 108.9,
        "f107": 111.0,
        "Ap": 5,
        "glat": 67.11,
        "glon": 212.95,
        "x2dist": 200e3,
        "x3dist": 100e3,
        "alt_min": 80e3,
        "alt_max": 1000e3,
        "alt_scale": [13.75e3, 20e3, 200e3, 200e3],
        "lx2": 40,
        "lx3": 1,
        "Bincl": 90,
        "Nmf": 5e11,
        "Nme": 2e11,
    }
    model.config(P, tmp_path)


def test_datetime_range():
    # datetime_range is a closed interval
    t = config.datetime_range(datetime(2012, 1, 1), datetime(2010, 1, 1), timedelta(seconds=1))
    assert isinstance(t, list)
    assert len(t) == 0

    t = config.datetime_range(datetime(2012, 1, 1), datetime(2012, 1, 1), timedelta(seconds=1))
    assert t == [datetime(2012, 1, 1)]

    t = config.datetime_range(datetime(2012, 1, 1), datetime(2012, 1, 1), timedelta(seconds=-1))
    assert t == [datetime(2012, 1, 1)]

    # this is how pandas.date_range works
    t = config.datetime_range(
        datetime(2012, 1, 1),
        datetime(2012, 1, 1, 0, 0, 1, microsecond=500000),
        timedelta(seconds=1),
    )
    assert t == [datetime(2012, 1, 1, 0, 0, 0), datetime(2012, 1, 1, 0, 0, 1)]

    # this is how pandas.date_range works
    t = config.datetime_range(
        datetime(2012, 1, 1, 0, 0, 0), datetime(2012, 1, 1, 0, 1, 30), timedelta(minutes=1)
    )
    assert t == [datetime(2012, 1, 1, 0, 0, 0), datetime(2012, 1, 1, 0, 1, 0)]


def test_no_config(tmp_path):
    p = read.config(tmp_path)
    assert p == {}


def test_no_nml(tmp_path):
    p = config.read_nml(tmp_path)
    assert p == {}


def test_no_ini(tmp_path):
    p = config.read_ini(tmp_path)
    assert p == {}


def test_nml_bad(tmp_path):
    blank = tmp_path / "foo"
    blank.touch()
    with pytest.raises(KeyError):
        config.parse_namelist(blank, "base")

    blank.write_text(
        """
&base
 t0 =
/
"""
    )
    with pytest.raises(KeyError):
        config.parse_namelist(blank, "base")


@pytest.mark.parametrize("group", ["base", ("base", "flags", "files", "precip", "efield")])
def test_namelist_exists(group):

    with importlib.resources.path("gemini3d.tests.config", "config_example.nml") as cfn:
        assert config.namelist_exists(cfn, "base")


@pytest.mark.parametrize("namelist", ["base", "flags", "files", "precip", "efield"])
def test_nml_namelist(namelist):

    with importlib.resources.path("gemini3d.tests.config", "config_example.nml") as cfn:
        params = config.parse_namelist(cfn, namelist)
    if "base" in namelist:
        assert params["time"][0] == datetime(2013, 2, 20, 5)

    if "files" in namelist:
        assert params["file_format"] == "h5"

    if "precip" in namelist:
        assert params["dtprec"] == timedelta(seconds=5)

    if "efield" in namelist:
        assert params["dtE0"] == timedelta(seconds=1)


def test_read_config_nml():

    with importlib.resources.path("gemini3d.tests.config", "config_example.nml") as cfn:
        params = read.config(cfn)
    assert params["time"][0] == datetime(2013, 2, 20, 5)


def test_read_config_ini():

    with importlib.resources.path("gemini3d.tests.config", "config_example.ini") as cfn:
        params = read.config(cfn)
    assert params["time"][0] == datetime(2013, 2, 20, 5)



#!/usr/bin/env python3
"""
run test
"""

import argparse
import sys
import subprocess
from pathlib import Path
import shutil

import gemini3d.web
import gemini3d.read
from . import mpi


def cli():
    p = argparse.ArgumentParser()
    p.add_argument("testname", help="name of test")
    p.add_argument("-mpiexec", help="mpiexec path")
    p.add_argument("exe", help="Gemini.bin executable binary")
    p.add_argument("outdir", help="output directory")
    p.add_argument("refdir", help="reference directory")
    p.add_argument("-np", help="force number of MPI images", type=int)
    p.add_argument(
        "-out_format",
        help="override config.nml output file format",
        choices=["h5", "nc", "raw"],
    )
    p.add_argument("-dryrun", help="only run first time step", action="store_true")
    P = p.parse_args()

    runner(
        P.testname,
        P.mpiexec,
        P.exe,
        P.outdir,
        P.refdir,
        mpi_count=P.np,
        out_format=P.out_format,
        dryrun=P.dryrun,
    )


def runner(
    test_name: str,
    mpiexec: str,
    exe: str,
    outdir: Path,
    refdir: Path,
    *,
    mpi_count: int = None,
    out_format: str = None,
    dryrun: bool = False,
):
    """configure and run a test
    This is usually called from CMake Ctest
    """

    outdir = Path(outdir).expanduser().resolve()
    refdir = Path(refdir).expanduser().resolve()

    ref = gemini3d.web.download_and_extract(test_name, refdir)

    # prepare simulation output directory
    input_dir = outdir / "inputs"
    input_dir.mkdir(parents=True, exist_ok=True)

    # a normal, non-test simulation already has all these files in the
    # output directory. here, we use a reference simulation input data.
    # the input data generation is tested elsewhere in PyGemini.
    # Here, we want to test that we can create
    # data match to reference outputs from reference inputs.

    if not (input_dir / "config.nml").is_file():
        shutil.copy2(ref / "inputs/config.nml", input_dir)

    cfg = gemini3d.read.config(ref)
    if not cfg:
        raise FileNotFoundError(f"{ref} does not appear to contain a config.nml file")

    # delete previous test run data to avoid restarting milestone and failing test
    if (outdir / "output.nml").is_file():
        stem = cfg["time"][0].strftime("%Y%m%d")
        for f in outdir.glob(f"{stem}*.h5"):
            f.unlink()

    # copy remaining input files needed
    if not (input_dir / cfg["indat_size"]).is_file():
        shutil.copy2(ref / cfg["indat_size"], input_dir)
        shutil.copy2(ref / cfg["indat_grid"], input_dir)
        shutil.copy2(ref / cfg["indat_file"], input_dir)
    if "precdir" in cfg and not (outdir / cfg["precdir"]).is_dir():
        shutil.copytree(ref / cfg["precdir"], outdir / cfg["precdir"])
    if "E0dir" in cfg and not (outdir / cfg["E0dir"]).is_dir():
        shutil.copytree(ref / cfg["E0dir"], outdir / cfg["E0dir"])
    if "neutral_perturb" in cfg and not (outdir / cfg["sourcedir"]).is_dir():
        shutil.copytree(ref / cfg["sourcedir"], outdir / cfg["sourcedir"])

    if not mpi_count:
        mpi_count = mpi.count(ref / cfg["indat_size"], 0)

    # have to get exe as absolute path
    exe_abs = Path(exe).resolve()
    if mpiexec:
        cmd = [mpiexec, "-np", str(mpi_count), str(exe_abs), str(outdir)]
    else:
        cmd = [str(exe_abs), str(outdir)]
    if out_format:
        cmd += ["-out_format", out_format]
    if dryrun:
        cmd.append("-dryrun")

    print(" ".join(cmd))

    ret = subprocess.run(cmd)
    if ret.returncode == 0:
        print("OK:", test_name)
    else:
        print("FAIL:", test_name, file=sys.stderr)

    raise SystemExit(ret.returncode)


if __name__ == "__main__":
    cli()



import argparse

from . import plot_all, grid, Efield, precip


def cli():

    p = argparse.ArgumentParser()
    p.add_argument("direc", help="directory to plot")
    p.add_argument(
        "which",
        help="which plots to make",
        choices=["all", "grid", "Efield", "precip"],
        nargs="+",
        default="all",
    )
    p.add_argument("--mayavi", help="do 3D Mayavi plots", action="store_true")
    p.add_argument("-var", help="plot these variables", nargs="+")
    p.add_argument(
        "-save", help="save plot format", choices=["png", "svg", "eps", "pdf"], default="png"
    )
    p = p.parse_args()

    if p.mayavi:
        from . import plot_3d

        plot_3d(p.direc, p.var, saveplot_fmt=p.save)
        return

    if "all" in p.which:
        plot_all(p.direc, p.var, saveplot_fmt=p.save)
    if "grid" in p.which:
        grid(p.direc)
    if "Efield" in p.which:
        Efield(p.direc)
    if "precip" in p.which:
        precip(p.direc)


if __name__ == "__main__":
    cli()



from __future__ import annotations
import os
import subprocess
import binascii
from pathlib import Path
import shutil
import importlib.resources


def hpc_submit_job(batcher: str, job_file: Path):
    """
    submits batch job
    """

    if batcher == "qsub":
        subprocess.run(["qsub", str(job_file)])
    else:
        raise LookupError(f"batcher {batcher} not yet listed. Please raise a Gemini Github Issue.")


def hpc_batch_create(batcher: str, out_dir: Path, cmd: list[str]) -> Path:
    """
    creates HPC batch scripts for known systems

    assumes that user-specific parameters like account number are already set
    as environment variables
    or static configuration files not handled by this scripts.

    This function assumes a script template exists, and it merely appends lines
    to the end of that template.

    TODO:

    1. determine estimated wallclock time to request on HPC
    2. determine requested HPC RAM per node (limit is main node)
    3. format number of nodes request
    """

    Nchar = 6  # arbitrary number of characters

    if batcher == "qsub":
        template = importlib.resources.read_text(__package__, "templates/qsub_template.job")
        job_file = out_dir / f"{binascii.b2a_hex(os.urandom(Nchar)).decode('ascii')}.job"
        print("writing job file", job_file)
        text = template + "\n" + " ".join(cmd)
        job_file.write_text(text)
    else:
        raise LookupError(f"batcher {batcher} not yet listed. Please raise a Gemini Github Issue.")

    return job_file


def hpc_batch_detect() -> str:
    """
    Assuming a known job batching system, we will create a template for the user
    to verify and then the user will run.
    """

    batcher = None

    if shutil.which("qsub"):
        batcher = "qsub"

    return batcher



import importlib.resources
import pytest
from datetime import datetime
from pathlib import Path

import gemini3d
import gemini3d.find as find
import gemini3d.web


def test_config(tmp_path):
    assert find.config(tmp_path) is None
    assert find.config(tmp_path / "not_exist") is None

    with importlib.resources.path("gemini3d.tests.config", "config_example.nml") as cfn:
        fn = find.config(cfn)
        assert fn == cfn


@pytest.mark.parametrize("name", ["2dew_fang"])
def test_grid(name, tmp_path):
    assert find.grid(tmp_path) is None
    assert find.grid(tmp_path / "not_exist") is None

    try:
        test_dir = gemini3d.web.download_and_extract(name, gemini3d.PYGEMINI_ROOT / "tests/data")
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    fn = find.grid(test_dir)
    assert fn.name.endswith("simgrid.h5")


@pytest.mark.parametrize("name", ["2dew_fang"])
def test_simsize(name, tmp_path):
    assert find.simsize(tmp_path) is None
    assert find.simsize(tmp_path / "not_exist") is None

    R = Path(gemini3d.__path__[0])

    try:
        test_dir = gemini3d.web.download_and_extract(name, R / "tests/data")
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    fn = find.simsize(test_dir)
    assert fn.name == "simsize.h5"


@pytest.mark.parametrize("name", ["2dew_fang"])
def test_frame(name, tmp_path):

    t = datetime(2013, 2, 20, 5)

    assert find.frame(tmp_path, t) is None
    assert find.frame(tmp_path / "not_exist", t) is None

    R = Path(gemini3d.__path__[0])

    try:
        test_dir = gemini3d.web.download_and_extract(name, R / "tests/data")
    except ConnectionError as e:
        pytest.skip(f"failed to download reference data {e}")

    fn = find.frame(test_dir, t)
    assert fn.name == "20130220_18000.000001.h5"



"""
functions for interfacing with Matlab or using Matlab data formats
"""

from __future__ import annotations
from pathlib import Path
import typing as T
import scipy.io


def grid(fn: Path, shape: bool = False) -> dict[str, T.Any]:
    """
    get simulation grid

    Parameters
    ----------
    fn: pathlib.Path
        filepath to simgrid file

    Returns
    -------
    grid: dict
        grid parameters
    """

    if shape:
        raise NotImplementedError("shape with Matlab")

    grid = scipy.io.loadmat(fn, squeeze_me=True)

    return grid


def simsize(path: Path) -> tuple[int, ...]:
    """
    get simulation size
    """
    path = Path(path).expanduser().resolve()

    f = scipy.io.loadmat(path, squeeze_me=True)
    if "lxs" in f:
        lxs = f["lxs"]
    elif "lx" in f:
        lxs = f["lx"]
    elif "lx1" in f:
        if f["lx1"].ndim > 0:
            lxs = (
                f["lx1"],
                f["lx2"],
                f["lx3"],
            )
        else:
            lxs = (f["lx1"], f["lx2"], f["lx3"])
    else:
        raise KeyError(f"could not find '/lxs', '/lx' or '/lx1' in {path.as_posix()}")

    return lxs


def state(fn: Path) -> dict[str, T.Any]:
    """
    load initial condition data
    """

    return scipy.io.loadmat(fn, squeeze_me=True)


def precip(fn: Path) -> dict[str, T.Any]:
    """
    load precipitation data
    """

    return scipy.io.loadmat(fn, squeeze_me=True)



"""
struct manpage:
https://docs.python.org/3/library/struct.html#struct-format-strings
"""

from __future__ import annotations
import numpy as np
from pathlib import Path
from datetime import datetime
import typing as T
import xarray

from .config import read_ini, read_nml
from . import find
from . import matlab
from . import LSP

from .raw import read as raw_read
from .hdf5 import read as h5read
from .nc4 import read as ncread


# do NOT use lru_cache--can have weird unexpected effects with complicated setups
def config(path: Path) -> dict[str, T.Any]:
    """
    read simulation input configuration

    .nml is strongly preferred, .ini is legacy.

    Parameters
    ----------
    path: pathlib.Path
        config file path

    Returns
    -------
    params: dict
        simulation parameters from config file
    """

    file = find.config(path)
    if not file:
        return {}  # {} instead of None to work with .get()

    if file.suffix == ".ini":
        P = read_ini(file)
    else:
        P = read_nml(file)

    return P


def simsize(path: Path) -> tuple[int, ...]:
    """ get simulation dimensions """

    fn = find.simsize(path)
    if not fn:
        return None

    if fn.suffix == ".h5":
        return h5read.simsize(fn)
    elif fn.suffix == ".nc":
        return ncread.simsize(fn)
    elif fn.suffix == ".dat":
        return raw_read.simsize(fn)
    elif fn.suffix == ".mat":
        return matlab.simsize(fn)
    else:
        raise ValueError("unkonwn simsize file type")


def grid(
    path: Path, *, var: list[str] = None, file_format: str = None, shape: bool = False
) -> dict[str, T.Any]:
    """
    get simulation grid

    Parameters
    ----------

    path: pathlib.Path
        path to simgrid.*
    var: list of str
        read only these grid variables
    file_format: str, optional
        force .h5, .nc, .dat (debugging)
    shape: bool, optional
        read only the shape of the grid instead of the data iteslf
    """

    fn = find.grid(path)
    if not fn:
        return {}

    if not file_format:
        file_format = fn.suffix[1:]

    if file_format == "dat":
        xg = raw_read.grid(fn.with_suffix(".dat"), shape=shape)
    elif file_format == "h5":
        xg = h5read.grid(fn.with_suffix(".h5"), var=var, shape=shape)
    elif file_format == "nc":
        xg = ncread.grid(fn.with_suffix(".nc"), var=var, shape=shape)
    elif file_format == "mat":
        xg = matlab.grid(fn.with_suffix(".mat"), shape=shape)
    else:
        raise ValueError(f"Unknown file type {fn}")

    xg["filename"] = fn

    return xg


def data(
    fn: Path,
    var: tuple[str, ...] | list[str] = None,
    *,
    file_format: str = None,
    cfg: dict[str, T.Any] = None,
    E0dir: Path = None,
) -> xarray.Dataset:
    """
    knowing the filename for a simulation time step, read the data for that time step

    Parameters
    ----------
    fn: pathlib.Path
        filename for this timestep
    file_format: str
        specify file extension of data files
    cfg: dict
        to avoid reading config.nml
    E0dir: pathlib.Path
        E0 directory

    Returns
    -------
    dat: xarray.Dataset
        simulation outputs
    """

    if not fn:
        return {}

    if not var:
        var = ["ne", "Ti", "Te", "v1", "v2", "v3", "J1", "J2", "J3", "Phi"]

    if isinstance(var, str):
        var = [var]

    fn = Path(fn).expanduser()
    fn_aurora = fn.parent / "aurmaps" / fn.name

    if not cfg:
        cfg = config(fn.parent)

    if not file_format:
        file_format = cfg["file_format"] if "file_format" in cfg else fn.suffix[1:]

    if file_format == "dat":
        flag = cfg.get("flagoutput")
        if flag == 0:
            dat = raw_read.frame3d_curvne(fn)
        elif flag == 1:
            dat = raw_read.frame3d_curv(fn)
        elif flag == 2:
            dat = raw_read.frame3d_curvavg(fn)
        else:
            raise ValueError(f"Unsure how to read {fn} with flagoutput {flag}")

        if fn_aurora.is_file():
            dat.update(raw_read.glow_aurmap(fn_aurora))

    elif file_format == "h5":
        flag = h5read.flagoutput(fn, cfg)

        if flag == 0:
            dat = h5read.frame3d_curvne(fn)
        elif flag == 1:
            dat = h5read.frame3d_curv(fn, var)
        elif flag == 2:
            dat = h5read.frame3d_curvavg(fn, var)
        else:
            raise ValueError(f"Unsure how to read {fn} with flagoutput {flag}")

        if fn_aurora.is_file():
            dat.update(h5read.glow_aurmap(fn_aurora))
    elif file_format == "nc":
        flag = ncread.flagoutput(fn, cfg)

        if flag == 0:
            dat = ncread.frame3d_curvne(fn)
        elif flag == 1:
            dat = ncread.frame3d_curv(fn, var)
        elif flag == 2:
            dat = ncread.frame3d_curvavg(fn, var)
        else:
            raise ValueError(f"Unsure how to read {fn} with flagoutput {flag}")

        if fn_aurora.is_file():
            dat.update(ncread.glow_aurmap(fn_aurora))
    else:
        raise ValueError(f"Unknown file type {fn}")

    lxs = simsize(fn.parent)

    # %% Derived variables
    if flag == 1:
        if {"ne", "v1", "Ti", "Te"}.intersection(var):
            dat["ne"] = (("x1", "x2", "x3"), dat["ns"][LSP - 1, :, :, :])
            # np.any() in case neither is an np.ndarray
            if dat["ns"].shape[0] != LSP or not np.array_equal(dat["ns"].shape[1:], lxs):
                raise ValueError(
                    f"may have wrong permutation on read. lxs: {lxs}  ns x1,x2,x3: {dat['ns'].shape}"
                )
        if "v1" in var:
            dat["v1"] = (
                ("x1", "x2", "x3"),
                (dat["ns"][:6, :, :, :] * dat["vs1"][:6, :, :, :]).sum(axis=0) / dat["ne"],
            )
        if "Ti" in var:
            dat["Ti"] = (
                ("x1", "x2", "x3"),
                (dat["ns"][:6, :, :, :] * dat["Ts"][:6, :, :, :]).sum(axis=0) / dat["ne"],
            )
        if "Te" in var:
            dat["Te"] = (("x1", "x2", "x3"), dat["Ts"][LSP - 1, :, :, :])

        if "J1" in var:
            # np.any() in case neither is an np.ndarray
            if np.any(dat["J1"].shape != lxs):
                raise ValueError("J1 may have wrong permutation on read")

    if "time" not in dat:
        dat.attrs["time"] = time(fn)

    if E0dir:
        fn_Efield = E0dir / fn.name
        if fn_Efield.is_file():
            dat.update(Efield(fn_Efield))

    return dat


def Efield(fn: Path, *, file_format: str = None) -> dict[str, T.Any]:
    """load Efield data "Efield_inputs"

    Parameters
    ----------
    fn: pathlib.Path
        filename for this timestep

    Returns
    -------
    dat: dict of np.ndarray
        electric field
    """

    if not fn:
        return {}

    fn = Path(fn).expanduser().resolve(strict=True)

    if not file_format:
        file_format = fn.suffix[1:]

    if file_format == "h5":
        E = h5read.Efield(fn)
    elif file_format == "nc":
        E = ncread.Efield(fn)
    elif file_format == "dat":
        E = raw_read.Efield(fn)
    else:
        raise ValueError(f"Unknown file type {fn}")

    return E


def precip(fn: Path, *, file_format: str = None) -> dict[str, T.Any]:
    """load precipitation to disk

    Parameters
    ----------
    fn: pathlib.Path
        path to precipitation file
    file_format: str
        file format to read

    Returns
    -------
    dat: dict
        precipitation
    """

    if not fn:
        return {}

    fn = Path(fn).expanduser().resolve(strict=True)

    if not file_format:
        file_format = fn.suffix[1:]

    if file_format == "h5":
        dat = h5read.precip(fn)
    elif file_format == "nc":
        dat = ncread.precip(fn)
    else:
        raise ValueError(f"unknown file format {file_format}")

    return dat


def frame(
    simdir: Path, time: datetime, *, var: list[str] = None, file_format: str = None
) -> xarray.Dataset:
    """
    load a frame of simulation data, automatically selecting the correct
    functions based on simulation parameters

    Parameters
    ----------
    simdir: pathlib.Path
        top-level directory of simulation output
    time: datetime.datetime
        time to load from simulation output
    file_format: str, optional
        "hdf5", "nc" for hdf5 or netcdf4 respectively

    Returns
    -------
    dat: xarray.Dataset
        simulation output for this time step
    """

    return data(find.frame(simdir, time, file_format), var=var, file_format=file_format)


def time(file: Path) -> datetime:
    """
    read simulation time of a file
    """

    if file.suffix == ".h5":
        t = h5read.time(file)
    elif file.suffix == ".nc":
        t = ncread.time(file)
    else:
        raise ValueError(f"unknown file format {file.suffix}")

    return t


def get_lxs(xg: dict[str, T.Any]) -> tuple[int, int, int]:

    lxs = None
    for k in ("lx", "lxs", "lx1"):
        if k in xg:
            if k == "lx1":
                lxs = [xg["lx1"], xg["lx2"], xg["lx3"]]
                break
            else:
                lxs = xg[k]

    if lxs is None:
        raise IndexError("Did not find grid size")

    return (lxs[0], lxs[1], lxs[2])



"""
generate, read, write Electric Field
"""

from __future__ import annotations
import xarray
import typing as T
import numpy as np
import logging
from scipy.special import erf

from . import write
from .config import datetime_range


def Efield_BCs(cfg: dict[str, T.Any], xg: dict[str, T.Any]) -> xarray.Dataset:
    """ generate E-field """

    # %% READ IN THE SIMULATION INFO
    lx1 = None
    lx2 = None
    lx3 = None
    for k in ("lx", "lxs"):
        if k in xg:
            lx1, lx2, lx3 = xg[k]
            break

    if lx1 is None:
        raise ValueError("size data not in Efield grid")

    # %% CREATE ELECTRIC FIELD DATASET
    # the Efield is interpolated from these, 100 x 100 is arbitrary
    llon = 100
    llat = 100
    # NOTE: cartesian-specific code
    if lx2 == 1:
        llon = 1
    elif lx3 == 1:
        llat = 1

    thetamin = xg["theta"].min()
    thetamax = xg["theta"].max()
    mlatmin = 90 - np.degrees(thetamax)
    mlatmax = 90 - np.degrees(thetamin)
    mlonmin = np.degrees(xg["phi"].min())
    mlonmax = np.degrees(xg["phi"].max())

    # add a 1% buff
    latbuf = 0.01 * (mlatmax - mlatmin)
    lonbuf = 0.01 * (mlonmax - mlonmin)

    E = xarray.Dataset(
        coords={
            "time": datetime_range(cfg["time"][0], cfg["time"][0] + cfg["tdur"], cfg["dtE0"]),
            "mlat": np.linspace(mlatmin - latbuf, mlatmax + latbuf, llat),
            "mlon": np.linspace(mlonmin - lonbuf, mlonmax + lonbuf, llon),
        }
    )

    Nt = E.time.size
    E.attrs["mlonmean"] = E.mlon.mean()
    E.attrs["mlatmean"] = E.mlat.mean()

    # %% WIDTH OF THE DISTURBANCE
    if "Efield_latwidth" in cfg:
        E.attrs["mlatsig"] = cfg["Efield_latwidth"] * (mlatmax - mlatmin)
        E.attrs["sigx3"] = cfg["Efield_latwidth"] * (xg["x3"].max() - xg["x3"].min())
    if "Efield_lonwidth" in cfg:
        E.attrs["mlonsig"] = cfg["Efield_lonwidth"] * (mlonmax - mlonmin)
        E.attrs["sigx2"] = cfg["Efield_lonwidth"] * (xg["x2"].max() - xg["x2"].min())

    # %% CREATE DATA FOR BACKGROUND ELECTRIC FIELDS
    # assign to zero in case not specifically assigned
    if "Exit" in cfg:
        E["Exit"] = (("time", "mlon", "mlat"), cfg["Exit"] * np.ones((Nt, llon, llat)))
    else:
        E["Exit"] = (("time", "mlon", "mlat"), np.zeros((Nt, llon, llat)))
    if "Eyit" in cfg:
        E["Eyit"] = (("time", "mlon", "mlat"), cfg["Eyit"] * np.ones((Nt, llon, llat)))
    else:
        E["Eyit"] = (("time", "mlon", "mlat"), np.zeros((Nt, llon, llat)))

    # %% CREATE DATA FOR BOUNDARY CONDITIONS FOR POTENTIAL SOLUTION

    # if 0 data is interpreted as FAC, else we interpret it as potential
    E["flagdirich"] = (("time",), np.zeros(Nt, dtype=np.int32))
    E["Vminx1it"] = (("time", "mlon", "mlat"), np.zeros((Nt, llon, llat)))
    E["Vmaxx1it"] = (("time", "mlon", "mlat"), np.zeros((Nt, llon, llat)))
    # these are just slices
    E["Vminx2ist"] = (("time", "mlat"), np.zeros((Nt, llat)))
    E["Vmaxx2ist"] = (("time", "mlat"), np.zeros((Nt, llat)))
    E["Vminx3ist"] = (("time", "mlon"), np.zeros((Nt, llon)))
    E["Vmaxx3ist"] = (("time", "mlon"), np.zeros((Nt, llon)))

    # %% synthesize feature
    if "Etarg" in cfg:
        E.attrs["Etarg"] = cfg["Etarg"]
        E = Efield_target(E, xg, lx1, lx2, lx3)
    elif "Jtarg" in cfg:
        E.attrs["Jtarg"] = cfg["Jtarg"]
        E = Jcurrent_target(E)
    else:
        # background only
        pass

    # %% check for NaNs
    # this is also done in Fortran, but just to help ensure results.
    check_finite(E["Exit"], "Exit")
    check_finite(E["Eyit"], "Eyit")
    check_finite(E["Vminx1it"], "Vminx1it")
    check_finite(E["Vmaxx1it"], "Vmaxx1it")
    check_finite(E["Vminx2ist"], "Vminx2ist")
    check_finite(E["Vmaxx2ist"], "Vmaxx2ist")
    check_finite(E["Vminx3ist"], "Vminx3ist")
    check_finite(E["Vmaxx3ist"], "Vmaxx3ist")

    # %% SAVE THESE DATA TO APPROPRIATE FILES
    # LEAVE THE SPATIAL AND TEMPORAL INTERPOLATION TO THE
    # FORTRAN CODE IN CASE DIFFERENT GRIDS NEED TO BE TRIED.
    # THE EFIELD DATA DO NOT TYPICALLY NEED TO BE SMOOTHED.
    write.Efield(E, cfg["E0dir"], cfg["file_format"])

    return E


def Jcurrent_target(E: xarray.Dataset) -> xarray.Dataset:

    S = (
        E["Jtarg"]
        * np.exp(-((E.mlon - E.mlonmean) ** 2) / 2 / E.mlonsig ** 2)
        * np.exp(-((E.mlat - E.mlatmean - 1.5 * E.mlatsig) ** 2) / 2 / E.mlatsig ** 2)
    )

    for t in E.time[6:]:
        E["flagdirich"].loc[t] = 0
        # could have different boundary types for different times
        E["Vmaxx1it"].loc[t] = S - E.Jtarg * np.exp(
            -((E.mlon - E.mlonmean) ** 2) / 2 / E.mlonsig ** 2
        ) * np.exp(-((E.mlat - E.mlatmean + 1.5 * E.mlatsig) ** 2) / 2 / E.mlatsig ** 2)

    return E


def Efield_target(
    E: xarray.Dataset, xg: dict[str, T.Any], lx1: int, lx2: int, lx3: int
) -> xarray.Dataset:
    """
    synthesize a feature
    """

    if E.Etarg > 1:
        logging.warning(f"Etarg units V/m -- is {E['Etarg']} V/m realistic?")

    # NOTE: h2, h3 have ghost cells, so we use lx1 instead of -1 to index
    # pk is a scalar.

    if lx3 == 1:
        # east-west
        S = E.Etarg * E.sigx2 * xg["h2"][lx1, lx2 // 2, 0] * np.sqrt(np.pi) / 2
        taper = erf((E.mlon - E.mlonmean) / E.mlonsig).values[:, None]
    elif lx2 == 1:
        # north-south
        S = E.Etarg * E.sigx3 * xg["h3"][lx1, 0, lx3 // 2] * np.sqrt(np.pi) / 2
        taper = erf((E.mlat - E.mlatmean) / E.mlatsig).values[None, :]
    else:
        # 3D
        S = E.Etarg * E.sigx2 * xg["h2"][lx1, lx2 // 2, 0] * np.sqrt(np.pi) / 2
        taper = (
            erf((E.mlon - E.mlonmean) / E.mlonsig).values[:, None]
            * erf((E.mlat - E.mlatmean) / E.mlatsig).values[None, :]
        )

    assert S.ndim == 0, "S is a scalar"

    for t in E.time:
        E["flagdirich"].loc[t] = 1
        E["Vmaxx1it"].loc[t] = S * taper

    return E


def check_finite(v: np.ndarray, name: str):

    i = ~np.isfinite(v)
    if i.any():
        raise ValueError(f"{np.count_nonzero(i)} NaN in {name} at {i.nonzero()}")



import os
import shutil
import subprocess
import functools
import typing as T
from pathlib import Path

Pathlike = T.Union[str, Path]


@functools.lru_cache()
def wsl_available() -> bool:
    """
    heuristic to detect if Windows Subsystem for Linux is available.

    Uses presence of /etc/os-release in the WSL image to say Linux is there.
    This is a de facto file standard across Linux distros.
    """

    has_wsl = False
    if os.name == "nt" and shutil.which("wsl"):
        has_wsl = wsl_file_exist("/etc/os-release")

    return has_wsl


def wsl_file_exist(file: Pathlike) -> bool:
    """
    path is specified as if in WSL
    NOT //wsl$/Ubuntu/etc/os-release
    but /etc/os-release
    """
    if os.name != "nt":
        return False

    try:
        return subprocess.run(["wsl", "test", "-f", str(file)], timeout=10).returncode == 0
    except subprocess.TimeoutExpired:
        return False



"""
unit tests of non-Fortran modules
"""

import pytest
from pytest import approx
import numpy as np
import math
import xarray
from datetime import datetime

import gemini3d.mpi as gm
import gemini3d.grid as grid
import gemini3d.coord as coord
import gemini3d.namelist as namelist
from gemini3d.utils import to_datetime

pi = math.pi


def test_to_datetime():
    x = xarray.Dataset(coords={"time": [datetime(2000, 1, 1)]})
    assert isinstance(to_datetime(x.time), datetime)

    assert isinstance(to_datetime(np.datetime64("2000-01-01")), datetime)

    assert isinstance(to_datetime(datetime(2000, 1, 1)), datetime)


def test_write_nml(tmp_path):
    dat = {"float": 42.0, "list_float": list(range(5)), "list_str": ["hi", "there"]}

    file = tmp_path / "test.nml"
    nml = "test"

    namelist.write(file, nml, dat)

    dat2 = namelist.read(file, nml)

    for k, v in dat.items():
        if isinstance(v, (tuple, list)):
            for i, w in enumerate(v):
                assert w == dat2[k][i]
        else:
            assert v == dat2[k]


def test_grid1d():
    x = grid.grid1d(100.0, 5)
    assert x == approx(np.arange(-100, 125, 25.0), rel=1e-6, abs=1e-8)

    x = grid.grid1d(100.0, 5, [200, 0.5, 9.5, 10])
    assert x == approx(
        [-50.25, -40.25, -30.25, -20.25, -10.25, -0.25, 0.25, 10.25, 20.25, 30.25, 40.25, 50.25],
        rel=1e-6,
        abs=1e-8,
    )


@pytest.mark.parametrize(
    "size,N,M",
    [
        ((None, 1, 40), 5, 5),
        ((None, 40, 1), 5, 5),
        ((None, 1, 40), 6, 5),
        ((None, 40, 1), 6, 5),
        ((None, 1, 40), 8, 8),
        ((None, 1, 40), 18, 10),
        ((None, 1, 40), 64, 40),
        ((None, 40, 1), 8, 8),
        ((None, 40, 1), 28, 20),
        ((None, 12, 8), 8, 8),
        ((None, 40, 36), 28, 24),
        ((None, 44, 54), 28, 27),
        ((None, 54, 44), 28, 27),
        ((None, 54, 44), 96, 88),
        ((None, 54, 44), 128, 108),
        ((None, 54, 44), 256, 216),
        ((None, 54, 44), 512, 396),
        ((None, 54, 44), 1024, 792),
    ],
)
def test_max_mpi(size, N, M):
    assert gm.max_mpi(size, N) == M


def test_coord():
    lat, lon = coord.geomag2geog(pi / 2, pi / 2)
    assert [lat, lon] == approx([0, 19], abs=1e-6, rel=0.001)

    theta, phi = coord.geog2geomag(0, 0)
    assert [theta, phi] == approx([1.50863496978059, 1.24485046147953], abs=1e-6, rel=0.001)

    alt, lon, lat = coord.UEN2geog(0, 0, 0, pi / 2, pi / 2)
    assert [alt, lat, lon] == approx([0, 0, 19], abs=1e-6, rel=0.001)

    z, x, y = coord.geog2UEN(0, 0, 0, pi / 2, pi / 2)
    assert [z, x, y] == approx([0, -2076275.16205889, 395967.844181141], abs=1e-6, rel=0.001)



from __future__ import annotations
from datetime import datetime
import numpy as np
import math
import xarray
import scipy.interpolate as interp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.ticker import MultipleLocator
import typing as T

from ..utils import git_meta
from ..read import get_lxs


if T.TYPE_CHECKING:
    import matplotlib.axes as mpla

mpl.rcParams["axes.formatter.limits"] = (-3, 4)
mpl.rcParams["axes.formatter.useoffset"] = False
mpl.rcParams["axes.formatter.min_exponent"] = 4

R_EARTH = 6370e3
REF_ALT = 300  # km

CB_LBL = {
    "ne": "$n_e (m^{-3})$",
    "v1": "$v_1 (ms^{-1})$",
    "Ti": "$T_i$ (K)",
    "Te": "$T_e$ (K)",
    "J1": "$J_1 (Am^{-2})$",
    "v2": "$v_2 (ms^{-1})$",
    "v3": "$v_3 (ms^{-1})$",
    "J2": "$J_2 (Am^{-2})$",
    "J3": "$J_3 (Am^{-2})$",
    "Phitop": r"$\Phi_{top}$ (V)",
    "Vmaxx1it": r"(V)",
}


def grid2plotfun(xg: dict[str, np.ndarray]):
    plotfun = None
    h1 = xg.get("h1")

    lxs = get_lxs(xg)

    if h1 is not None:
        minh1 = h1.min()
        maxh1 = h1.max()
        if (abs(minh1 - 1) > 1e-4) or (abs(maxh1 - 1) > 1e-4):  # curvilinear grid
            if (lxs[1] > 1) and (lxs[2] > 1):
                plotfun = plot3D_curv_frames_long
            else:
                plotfun = plot2D_curv
    if plotfun is None:  # cartesian grid
        if (lxs[1] > 1) and (lxs[2] > 1):
            plotfun = plot3D_cart_frames_long_ENU
        else:
            plotfun = plot2D_cart

    return plotfun


def plot3D_curv_frames_long(
    time: datetime,
    grid: dict[str, np.ndarray],
    parm: np.ndarray,
    name: str,
    fg: Figure = None,
    **kwargs,
):
    raise NotImplementedError


def plot2D_curv(
    time: datetime,
    grid: dict[str, np.ndarray],
    parm: np.ndarray,
    name: str,
    fg: Figure = None,
    **kwargs,
):
    raise NotImplementedError


def plot12(
    x: np.ndarray,
    z: np.ndarray,
    parm: np.ndarray,
    name: str,
    cmap: str,
    vmin: float,
    vmax: float,
    fg: Figure = None,
    ax: "mpla.Axes" = None,
) -> Figure:

    if parm.ndim != 2:
        raise ValueError(f"data must have 2 dimensions, you have {parm.shape}")

    if fg is None:
        fg = Figure(constrained_layout=True)
    if ax is None:
        ax = fg.gca()

    hi = ax.pcolormesh(x / 1e3, z / 1e3, parm, cmap=cmap, vmin=vmin, vmax=vmax, shading="nearest")
    ax.yaxis.set_major_locator(MultipleLocator(100))
    ax.set_xlabel("eastward dist. (km)")
    ax.set_ylabel("upward dist. (km)")
    ax.axhline(REF_ALT, color="w", linestyle="--", linewidth=2)
    fg.colorbar(hi, ax=ax, label=CB_LBL[name])

    return fg


def plot13(
    y: np.ndarray,
    z: np.ndarray,
    parm: np.ndarray,
    name: str,
    cmap: str,
    vmin: float,
    vmax: float,
    fg: Figure = None,
    ax: "mpla.Axes" = None,
) -> Figure:

    if parm.ndim != 2:
        raise ValueError(f"data must have 2 dimensions, you have {parm.shape}")

    if fg is None:
        fg = Figure(constrained_layout=True)
    if ax is None:
        ax = fg.gca()

    hi = ax.pcolormesh(y / 1e3, z / 1e3, parm, cmap=cmap, vmin=vmin, vmax=vmax, shading="nearest")
    ax.yaxis.set_major_locator(MultipleLocator(100))
    ax.set_xlabel("northward dist. (km)")
    ax.set_ylabel("upward dist. (km)")
    fg.colorbar(hi, ax=ax, label=CB_LBL[name])

    return fg


def plot23(
    x: np.ndarray,
    y: np.ndarray,
    parm: np.ndarray,
    name: str,
    cmap: str,
    vmin: float,
    vmax: float,
    fg: Figure = None,
    ax: "mpla.Axes" = None,
) -> Figure:

    if parm.ndim != 2:
        raise ValueError(f"data must have 2 dimensions, you have {parm.shape}")

    if fg is None:
        fg = Figure(constrained_layout=True)
    if ax is None:
        ax = fg.gca()

    hi = ax.pcolormesh(x / 1e3, y / 1e3, parm, cmap=cmap, vmin=vmin, vmax=vmax, shading="nearest")
    ax.set_xlabel("eastward dist. (km)")
    ax.set_ylabel("northward dist. (km)")
    fg.colorbar(hi, ax=ax, label=CB_LBL[name])

    return fg


def plot1d2(
    x: np.ndarray, parm: np.ndarray, name: str, fg: Figure = None, ax: "mpla.Axes" = None
) -> Figure:

    if parm.ndim != 1:
        raise ValueError("expecting 1-D data oriented east-west (along latitude)")

    if fg is None:
        fg = Figure(constrained_layout=True)
    if ax is None:
        ax = fg.gca()

    ax.plot(x / 1e3, parm)
    ax.set_xlabel("eastward dist. (km)")
    ax.set_ylabel(CB_LBL[name])

    return fg


def plot1d3(
    y: np.ndarray, parm: np.ndarray, name: str, fg: Figure = None, ax: "mpla.Axes" = None
) -> Figure:

    if parm.ndim != 1:
        raise ValueError("expecting 1-D data oriented east-west (along latitude)")

    if fg is None:
        fg = Figure(constrained_layout=True)
    if ax is None:
        ax = fg.gca()

    ax.plot(y / 1e3, parm)
    ax.set_xlabel("northward dist. (km)")
    ax.set_ylabel(CB_LBL[name])

    return fg


def plot_interp(
    time: datetime,
    grid: dict[str, np.ndarray],
    parm: xarray.DataArray,
    name: str,
    fg: Figure = None,
    **kwargs,
) -> Figure:

    """

    Parameters
    ----------

    xp:  eastward distance (rads.)
        should be interpreted as northward distance (in rads.).
        Irrespective of ordering of xg.theta, this will be monotonic increasing!!!

    zp: altitude (meters)

    y:  this is a mag colat. coordinate and is only used for defining
        grid in linspaces below, runs backward from north distance,
        hence the negative sign
    """

    if fg is None:
        fg = Figure(constrained_layout=True)

    meta = git_meta()

    cmap = None
    is_Efield = False
    vmin = None
    vmax = None
    if name.startswith("J") or name == "Phitop":
        cmap = "bwr"
        vmax = abs(parm).max()
        vmin = -vmax
    elif name.startswith("v"):
        cmap = "bwr"
        vmax = 80.0
        vmin = -vmax
    elif name.startswith(("V", "E")):
        is_Efield = True
        cmap = "bwr"
        vmax = abs(parm).max()
        vmin = -vmax
    elif name.startswith("T"):
        vmin = 0.0
        vmax = parm.max()
    elif name.startswith("n"):
        vmin = 1e-7

    # %% SIZE OF SIMULATION
    lxs = get_lxs(grid)

    lx1, lx2, lx3 = lxs
    inds1 = slice(2, lx1 + 2)
    inds2 = slice(2, lx2 + 2)
    inds3 = slice(2, lx3 + 2)
    # %% SIZE OF PLOT GRID THAT WE ARE INTERPOLATING ONTO
    meantheta = grid["theta"].mean()
    # this is a mag colat. coordinate and is only used for defining grid in linspaces below
    # runs backward from north distance, hence the negative sign
    # [radians]
    y = -1 * (grid["theta"] - meantheta)
    # eastward distance [radians]
    x = grid["x2"][inds2] / R_EARTH / math.sin(meantheta)
    # altitude [meters]
    z = grid["alt"] / 1e3

    # arbitrary output plot resolution
    lxp = 500
    lyp = 500
    lzp = 500

    # eastward distance [meters]
    xp = np.linspace(x.min(), x.max(), lxp) * R_EARTH * math.sin(meantheta)
    # northward distance [meters]
    yp = np.linspace(y.min(), y.max(), lyp) * R_EARTH
    # upward distance [meters]
    zp = np.linspace(z.min(), z.max(), lzp) * 1e3

    # %% INTERPOLATE ONTO PLOTTING GRID
    if lxs[2] == 1:  # alt./lon. slice
        ax = fg.gca()
        ax.set_title(f"{name}: {time.isoformat()}  {meta['commit']}")
        # meridional meshgrid, this defines the grid for plotting
        # slice expects the first dim. to be "y" ("z" in the 2D case)
        # %% CONVERT ANGULAR COORDINATES TO MLAT,MLON
        i = np.argsort(xp)  # FIXME: this was in Matlab code--what is its purpose?

        if name == "rayleighs":
            f = interp.interp1d(grid["x2"][inds2], parm, axis=1, bounds_error=False)
            # hack for pcolormesh to put labels in center of pixel
            wl = kwargs["wavelength"] + [""]
            hi = ax.pcolormesh(xp / 1e3, np.arange(len(wl)), f(xp)[:, i], shading="nearest")
            ax.set_yticks(np.arange(len(wl)) + 0.5)
            ax.set_yticklabels(wl)
            ax.set_ylim(0, len(wl) - 1)
            fg.colorbar(hi, ax=ax, aspect=60, pad=0.01)
            # end hack
            ax.set_ylabel(r"wavelength $\AA$")
            ax.set_xlabel("eastward dist. (km)")
        elif parm.ndim == 2:
            f = interp.interp2d(grid["x2"][inds2], grid["x1"][inds1], parm, bounds_error=False)
            plot12(xp[i], zp, f(xp, zp)[:, i], name, cmap, vmin, vmax, fg, ax)
        elif parm.ndim == 1:  # phitop
            f = interp.interp1d(grid["x2"][inds2], parm, bounds_error=False)
            plot1d2(xp, f(xp), name, fg, ax)
        else:
            raise ValueError(f"{name}: only 2D and 1D data are expected--squeeze data")
    elif lxs[1] == 1:  # alt./lat. slice
        ax = fg.gca()
        ax.set_title(f"{name}: {time.isoformat()}  {meta['commit']}")
        # so north dist, east dist., alt.
        # slice expects the first dim. to be "y"
        # %% CONVERT ANGULAR COORDINATES TO MLAT,MLON
        i = np.argsort(yp)  # FIXME: this was in Matlab code--what is its purpose?

        if name == "rayleighs":
            # FIXME: this needs to be tested
            f = interp.interp1d(grid["x3"][inds3], parm, axis=1, bounds_error=False)
            # hack for pcolormesh to put labels in center of pixel
            wl = kwargs["wavelength"] + [""]
            hi = ax.pcolormesh(np.arange(len(wl)), yp / 1e3, f(yp)[:, i].T, shading="nearest")
            ax.set_xticks(np.arange(len(wl)) + 0.5)
            ax.set_xticklabels(wl)
            ax.set_xlim(0, len(wl) - 1)
            fg.colorbar(hi, ax=ax, aspect=60, pad=0.01)
            # end hack
            ax.set_xlabel(r"wavelength $\AA$")
            ax.set_ylabel("northward dist. (km)")
        elif parm.ndim == 2:
            f = interp.interp2d(grid["x3"][inds3], grid["x1"][inds1], parm, bounds_error=False)
            parmp = f(yp, zp).reshape((lzp, lyp))
            plot13(yp[i], zp, parmp[:, i], name, cmap, vmin, vmax, fg, ax)
        elif parm.ndim == 1:  # phitop
            f = interp.interp1d(grid["x3"][inds3], parm, bounds_error=False)
            plot1d3(yp, f(yp), name, fg, ax)
        else:
            raise ValueError(f"{name}: only 2D and 1D data are expected--squeeze data")

    else:  # 3-panel plot, vs. single-panel plots of 2-D cases
        if name == "rayleighs":
            bright_east_north(
                fg,
                grid,
                parm,
                xp,
                yp,
                inds2,
                inds3,
                cmap,
                vmin,
                vmax,
                name,
                time,
                kwargs["wavelength"],
            )
            return fg
        elif parm.ndim == 3:
            fg.set_size_inches((18, 5))
            axs = fg.subplots(1, 3, sharey=False, sharex=False)
            fg.suptitle(f"{name}: {time.isoformat()}  {meta['commit']}", y=0.99)
        elif is_Efield:
            # like phitop, SINGLE plot
            mag_lonlat(fg, grid, parm, cmap, vmin, vmax, name, time)
            return fg
        else:
            # like phitop, SINGLE plot
            east_north(fg, grid, parm, xp, yp, inds2, inds3, cmap, vmin, vmax, name, time)
            return fg

        # %% CONVERT TO DISTANCE UP, EAST, NORTH (left panel)
        # JUST PICK AN X3 LOCATION FOR THE MERIDIONAL SLICE PLOT,
        # AND AN ALTITUDE FOR THE LAT./LON. SLICE
        ix3 = lx3 // 2 - 1  # arbitrary slice, to match Matlab
        f = interp.interp2d(
            grid["x2"][inds2], grid["x1"][inds1], parm[:, :, ix3], bounds_error=False
        )
        # CONVERT ANGULAR COORDINATES TO MLAT,MLON
        ix = np.argsort(xp)
        iy = np.argsort(yp)
        plot12(xp[ix], zp, f(xp, zp)[:, ix], name, cmap, vmin, vmax, fg, axs[0])
        # %% LAT./LONG. SLICE COORDINATES (center panel)
        zp2 = REF_ALT
        X3, Y3, Z3 = np.meshgrid(xp, yp, zp2 * 1e3)
        # transpose: so north dist, east dist., alt.
        parmp = interp.interpn(
            points=(grid["x1"][inds1], grid["x2"][inds2], grid["x3"][inds3]),
            values=parm.values,
            xi=np.column_stack((Z3.ravel(), X3.ravel(), Y3.ravel())),
            bounds_error=False,
        ).reshape((1, lxp, lyp))

        parmp = parmp[:, :, iy]  # must be indexed in two steps
        plot23(xp[ix], yp[iy], parmp[0, ix, :], name, cmap, vmin, vmax, fg, axs[1])
        # %% ALT/LAT SLICE (right panel)
        ix2 = lx2 // 2 - 1  # arbitrary slice, to match Matlab
        f = interp.interp2d(
            grid["x3"][inds3], grid["x1"][inds1], parm[:, ix2, :], bounds_error=False
        )
        plot13(yp[iy], zp, f(yp, zp)[:, iy], name, cmap, vmin, vmax, fg, axs[2])

    return fg


plot3D_cart_frames_long_ENU = plot_interp
plot2D_cart = plot_interp


def bright_east_north(
    fg, grid, parm, xp, yp, inds2, inds3, cmap, vmin, vmax, name, time, wavelength
):
    meta = git_meta()

    axs = fg.subplots(2, 2, sharey=True, sharex=True).ravel()
    fg.suptitle(f"{name}: {time.isoformat()}  {meta['commit']}", y=0.99)
    # arbitrary pick of which emission lines to plot lat/lon slices
    for j, i in enumerate([1, 3, 4, 8]):
        f = interp.interp2d(grid["x3"][inds3], grid["x2"][inds2], parm[i, :, :], bounds_error=False)
        hi = axs[j].pcolormesh(xp / 1e3, yp / 1e3, f(yp, xp), shading="nearest")
        axs[j].set_title(wavelength[i] + r"$\AA$")
        fg.colorbar(hi, ax=axs[j], label="Rayleighs")
    axs[2].set_xlabel("eastward dist. (km)")
    axs[2].set_ylabel("northward dist. (km)")


def east_north(fg, grid, parm, xp, yp, inds2, inds3, cmap, vmin, vmax, name, time):
    meta = git_meta()

    ax = fg.gca()

    f = interp.interp2d(grid["x3"][inds3], grid["x2"][inds2], parm, bounds_error=False)
    hi = ax.pcolormesh(
        xp / 1e3, yp / 1e3, f(yp, xp), cmap=cmap, vmin=vmin, vmax=vmax, shading="nearest"
    )
    ax.set_xlabel("eastward dist. (km)")
    ax.set_ylabel("northward dist. (km)")
    ax.set_title(f"{name}: {time.isoformat()}  {meta['commit']}")
    fg.colorbar(hi, ax=ax, label=CB_LBL[name])


def mag_lonlat(fg, grid, parm, cmap, vmin, vmax, name, time):
    meta = git_meta()

    ax = fg.gca()
    hi = ax.pcolormesh(
        grid["mlon"], grid["mlat"], parm, cmap=cmap, vmin=vmin, vmax=vmax, shading="nearest"
    )
    ax.set_xlabel("magnetic longitude (deg.)")
    ax.set_ylabel("magnetic latitude (deg.)")
    ax.set_title(f"{name}: {time.isoformat()}  {meta['commit']}")
    fg.colorbar(hi, ax=ax, label=CB_LBL[name])



"""
test prereq build
"""

import pytest
import shutil
import os

import gemini3d.prereqs as gcp
import gemini3d.cmake as cmake


@pytest.mark.skipif(shutil.which("gfortran") is None, reason="test assumes GCC")
def test_find_gcc():
    comps = gcp.gcc_compilers()
    assert "gfortran" in comps["FC"]


@pytest.mark.skipif(shutil.which("ifort") is None, reason="test assumes Intel compilers")
def test_find_intel():
    comps = gcp.intel_compilers()
    assert "ifort" in comps["FC"]


@pytest.mark.parametrize("name", ["MPI COMPONENTS C Fortran", "LAPACK"])
def test_find_cmake(name):

    assert cmake.find_library(name, [], None), f"{name} not found"


def test_cmake_notfound():

    assert not cmake.find_library("abc123zxyb", [], None), "should not be found"


@pytest.mark.parametrize("name", ["lapack", "scalapack"])
def test_libs(name, tmp_path):
    """ test that exception isn't raised for dryrun """
    dirs = {"prefix": tmp_path / f"install/{name}", "workdir": tmp_path / f"build/{name}"}

    gcp.setup_libs(name, dirs, compiler="gcc", wipe=True, dryrun=True)


@pytest.mark.skipif(os.name != "nt", reason="these are windows-only tests")
@pytest.mark.parametrize("name", ["openmpi"])
def test_not_for_windows(name, tmp_path):

    dirs = {"prefix": tmp_path / f"install/{name}", "workdir": tmp_path / f"build/{name}"}

    with pytest.raises(NotImplementedError):
        gcp.setup_libs(name, dirs, compiler="gcc", wipe=True, dryrun=True)



from __future__ import annotations
import hashlib
import urllib.request
import urllib.error
import socket
import json
import zipfile
import tarfile
import typing as T
from pathlib import Path
import importlib.resources
import shutil
import subprocess

Pathlike = T.Union[str, Path]


def git_download(path: Path, repo: str, tag: str = None):
    """
    Use Git to download code repo.
    """

    git = shutil.which("git")

    if not git:
        raise FileNotFoundError("Git not found.")

    if not tag:
        if not path.is_dir():
            subprocess.check_call([git, "clone", repo, "--depth", "1", str(path)])
        return

    if path.is_dir():
        # don't use "git -C" for old HPC
        ret = subprocess.run([git, "checkout", tag], cwd=str(path))
        if ret.returncode != 0:
            ret = subprocess.run([git, "fetch"], cwd=str(path))
            if ret.returncode != 0:
                raise RuntimeError(f"could not fetch {path}  Maybe try removing this directory.")
            subprocess.check_call([git, "checkout", tag], cwd=str(path))
    else:
        # shallow clone
        subprocess.check_call([git, "clone", repo, "--branch", tag, "--single-branch", str(path)])


def download_and_extract(test_name: str, data_dir: Path) -> Path:

    with importlib.resources.path("gemini3d.tests", "gemini3d_url.json") as url_ini:
        z = get_test_params(test_name, url_ini, data_dir)

        if z["dir"].is_dir():
            return z["dir"]

        try:
            url_retrieve(z["url"], z["zip"], ("md5", z["md5"]))
        except (ConnectionError, ValueError) as e:
            raise ConnectionError(f"problem downloading reference data {e}")

        try:
            extract_zip(z["zip"], z["dir"])
        except zipfile.BadZipFile:
            # bad download, delete and try again (maybe someone hit Ctrl-C during download)
            z["zip"].unlink()
            url_retrieve(z["url"], z["zip"], ("md5", z["md5"]))
            extract_zip(z["zip"], z["dir"])

    return z["dir"]


def get_test_params(test_name: str, url_file: Path, ref_dir: Path) -> dict[str, T.Any]:
    """ get URL and MD5 for a test name """
    json_str = Path(url_file).expanduser().read_text()
    urls = json.loads(json_str)

    z = {
        "url": urls[test_name]["url"],
        "dir": ref_dir / f"test{test_name}",
        "zip": ref_dir / f"test{test_name}.zip",
    }

    if urls[test_name].get("md5"):
        z["md5"] = urls[test_name]["md5"]
    else:
        z["md5"] = None

    return z


def url_retrieve(
    url: str,
    outfile: Pathlike,
    filehash: tuple[str, str] | list[str] = None,
    overwrite: bool = False,
):
    """
    Parameters
    ----------
    url: str
        URL to download from
    outfile: pathlib.Path
        output filepath (including name)
    filehash: tuple of str, str
        hash type (md5, sha1, etc.) and hash
    overwrite: bool
        overwrite if file exists
    """
    outfile = Path(outfile).expanduser().resolve()
    if outfile.is_dir():
        raise ValueError("Please specify full filepath, including filename")
    # need .resolve() in case intermediate relative dir doesn't exist
    if overwrite or not outfile.is_file():
        outfile.parent.mkdir(parents=True, exist_ok=True)
        print(f"{url} => {outfile}")
        try:
            urllib.request.urlretrieve(url, str(outfile))
        except (socket.gaierror, urllib.error.URLError) as err:
            raise ConnectionError(f"could not download {url} due to {err}")

    if filehash and filehash[1]:
        if not file_checksum(outfile, filehash[0], filehash[1]):
            raise ValueError(f"Hash mismatch: {outfile}")


def file_checksum(fn: Path, mode: str, filehash: str) -> bool:
    h = hashlib.new(mode)
    h.update(fn.read_bytes())
    return h.hexdigest() == filehash


def extract_zip(fn: Pathlike, outpath: Pathlike, overwrite: bool = False):
    outpath = Path(outpath).expanduser().resolve()
    # need .resolve() in case intermediate relative dir doesn't exist
    if outpath.is_dir() and not overwrite:
        return

    fn = Path(fn).expanduser().resolve()
    with zipfile.ZipFile(fn) as z:
        z.extractall(str(outpath.parent))


def extract_tar(fn: Pathlike, outpath: Pathlike, overwrite: bool = False):
    outpath = Path(outpath).expanduser().resolve()
    # need .resolve() in case intermediate relative dir doesn't exist
    if outpath.is_dir() and not overwrite:
        return

    fn = Path(fn).expanduser().resolve()
    if not fn.is_file():
        # tarfile gives confusing error on missing file
        raise FileNotFoundError(fn)

    try:
        with tarfile.open(fn) as z:
            z.extractall(str(outpath.parent))
    except tarfile.TarError as e:
        raise RuntimeError(
            f"""failed to extract {fn} with error {e}.
This file may be corrupt or system libz may be broken.
Try deleting {fn} or manually extracting it."""
        )
